import tempfile
import os
import yaml
from jsonschema.exceptions import ValidationError
from yggdrasil import yamlfile
from yggdrasil.backwards import StringIO
from yggdrasil.tests import YggTestClass, assert_raises, assert_equal
_yaml_env = 'TEST_YAML_FILE'


def direct_translate(msg):  # pragma: no cover
    r"""Test translator that just returns passed message."""
    return msg


def test_load_yaml():
    r"""Test loading yaml."""
    cwd = os.getcwd()
    fname = os.path.join(cwd, 'test_load_yaml.yml')
    dict_write = {'test': 'hello'}
    dict_read = {'test': 'hello',
                 'working_dir': cwd}
    contents = yaml.dump(dict_write)
    with open(fname, 'w') as fd:
        yaml.dump(dict_write, fd)
    try:
        # Dictionary
        out = yamlfile.load_yaml(dict_write)
        assert_equal(out, dict_read)
        # File name
        out = yamlfile.load_yaml(fname)
        assert_equal(out, dict_read)
        # Open file object
        with open(fname, 'r') as fd:
            out = yamlfile.load_yaml(fd)
            assert_equal(out, dict_read)
        # Open stream
        out = yamlfile.load_yaml(StringIO(contents))
        assert_equal(out, dict_read)
    finally:
        # Remove file
        if os.path.isfile(fname):
            os.remove(fname)
            

def test_load_yaml_error():
    r"""Test error on loading invalid file."""
    assert_raises(IOError, yamlfile.load_yaml, 'invalid')


def test_parse_component_error():
    r"""Test errors in parse_component."""
    assert_raises(TypeError, yamlfile.parse_component,
                  1, 'invalid', 'invalid')
    assert_raises(ValueError, yamlfile.parse_component,
                  {}, 'invalid', 'invalid')


class YamlTestBase(YggTestClass):
    r"""Test base for yamlfile."""
    _contents = tuple()

    def __init__(self, *args, **kwargs):
        super(YamlTestBase, self).__init__(*args, **kwargs)
        self.files = []
        for i in range(self.nfiles):
            self.files.append(self.get_fname(i))

    @property
    def nfiles(self):
        r"""int: Number of files."""
        return len(self.contents)

    @property
    def contents(self):
        r"""tuple: Contents of files."""
        return self._contents

    @property
    def yaml_env(self):
        r"""str: Environment variable where file path is stored."""
        return _yaml_env

    def setup(self):
        r"""Write contents to temp file."""
        super(YamlTestBase, self).setup()
        if self.nfiles > 0:
            os.environ[self.yaml_env] = self.files[0]
        for fname, content in zip(self.files, self.contents):
            with open(fname, 'w') as f:
                f.write('\n'.join(content))

    def teardown(self):
        r"""Remove the temporary file if it exists."""
        for fname in self.files:
            if os.path.isfile(fname):
                os.remove(fname)
        super(YamlTestBase, self).teardown()

    def get_fname(self, idx=0):
        r"""Path to temporary file."""
        return os.path.join(tempfile.gettempdir(),
                            '%s_%s_%d.yml' % (
                                tempfile.gettempprefix(), self.uuid, idx))

    def create_instance(self):
        r"""Disabled: Create a new instance of the class."""
        return None

    def test_parse_yaml(self):
        r"""Test successfully reading & parsing yaml."""
        if self.nfiles == 0:
            pass
        elif self.nfiles == 1:
            yamlfile.parse_yaml(self.files[0])
        else:
            yamlfile.parse_yaml(self.files)


class YamlTestBaseError(YamlTestBase):
    r"""Test error for yamlfile."""
    _error = None

    def test_parse_yaml(self):
        r"""Test error reading & parsing yaml."""
        if (self._error is None) or (self.nfiles == 0):
            return
        assert_raises(self._error, yamlfile.parse_yaml, self.files)


class TestYamlModelOnly(YamlTestBase):
    r"""Test parsing of different numbers/styles of models."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c'],
                 ['model:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c'],
                 ['model:',
                  '  name: modelC',
                  '  driver: GCCModelDriver',
                  '  args: ./src/modelC.c'],
                 ['models:',
                  '  name: modelD',
                  '  driver: GCCModelDriver',
                  '  args: ./src/modelD.c'], )


class TestYamlServerClient(YamlTestBase):
    r"""Test specification of server/client models."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    is_server: True'],
                 ['model:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c',
                  '    client_of: modelA'],)


class TestYamlIODrivers(YamlTestBase):
    r"""Test full specification of IO drivers."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - name: inputA',
                  '        driver: FileInputDriver',
                  '        translator: %s:direct_translate' % __name__,
                  '        onexit: printStatus',
                  '        args: {{ %s }}' % _yaml_env,
                  '    outputs:',
                  '      - name: outputA',
                  '        driver: FileOutputDriver',
                  '        translator: %s:direct_translate' % __name__,
                  '        onexit: printStatus',
                  '        args: fileA.txt',
                  '      - name: outputA2',
                  '        driver: OutputDriver',
                  '        translator: %s:direct_translate' % __name__,
                  '        onexit: printStatus',
                  '        args: A_to_B'],
                 ['model:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c',
                  '    input:',
                  '      - name: inputB',
                  '        driver: FileInputDriver',
                  '        args: {{ %s }}' % _yaml_env,
                  '      - name: inputB2',
                  '        driver: InputDriver',
                  '        args: A_to_B'],
                 ['model:',
                  '  name: modelC',
                  '  driver: GCCModelDriver',
                  '  args: ./src/modelC.c',
                  '  input:',
                  '    name: inputC',
                  '    driver: FileInputDriver',
                  '    args: {{ %s }}' % _yaml_env],
                 ['models:',
                  '  name: modelD',
                  '  driver: GCCModelDriver',
                  '  args: ./src/modelD.c',
                  '  inputs:',
                  '    name: inputD',
                  '    driver: FileInputDriver',
                  '    args: {{ %s }}' % _yaml_env], )


class TestYamlConnection(YamlTestBase):
    r"""Test connection between I/O channels."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '',
                  'connections:',
                  '  - input: outputB',
                  '    output: inputA'],
                 ['models:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c',
                  '    outputs:',
                  '      - outputB'],)


class TestYamlIODatatype(YamlTestBase):
    r"""Test specification of datatype via schema."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - name: inputA',
                  '        type: object',
                  '        properties:',
                  '          a: int',
                  "          b: {'type': float, 'units': 'cm'}",
                  '',
                  'connections:',
                  "  - from: ['outputB:0', 'outputB:1']",
                  "    to: ['inputA:a', 'inputA:b']"],
                 ['models:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c',
                  '    outputs:',
                  '      - name: outputB',
                  '        type: [int, float]'],)


class TestYamlConnectionFork(YamlTestBase):
    r"""Test connection between I/O channels."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - inputs: ',
                  '      - outputB',
                  '      - outputC',
                  '    output: inputA',
                  '  - input: outputA',
                  '    outputs:',
                  '      - inputB',
                  '      - inputC'],
                 ['models:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c',
                  '    inputs:',
                  '      - inputB',
                  '    outputs:',
                  '      - outputB'],
                 ['models:',
                  '  - name: modelC',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelC.c',
                  '    inputs:',
                  '      - inputC',
                  '    outputs:',
                  '      - outputC'],)


class TestYamlConnectionTranslator(YamlTestBase):
    r"""Test connection between I/O channels."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '',
                  'connections:',
                  '  - input: outputB',
                  '    output: inputA',
                  '    translator: %s:direct_translate' % __name__],
                 ['models:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c',
                  '    outputs:',
                  '      - outputB'],)


class TestYamlConnectionInputFile(YamlTestBase):
    r"""Test connection with File."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input:',
                  '      - {{ %s }}' % _yaml_env,
                  '    read_meth: all',
                  '    output: inputA',
                  '  - input: outputA',
                  '    output:',
                  '      - output.txt',
                  '    write_meth: all'],)


class TestYamlConnectionInputFile_wait(YamlTestBase):
    r"""Test connection with File where wait_for_creation specified."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input:',
                  '      - {{ %s }}' % _yaml_env,
                  '    read_meth: all',
                  '    output: inputA',
                  '  - input: outputA',
                  '    output:',
                  '      - output.txt',
                  '    write_meth: all'],)


class TestYamlConnectionInputAsciiFile(YamlTestBase):
    r"""Test connection with AsciiFile."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input: invalid.txt',
                  '    output: inputA',
                  '    wait_for_creation: 4',
                  '    read_meth: line',
                  '  - input: outputA',
                  '    output: output.txt',
                  '    write_meth: line'],)


class TestYamlConnectionInputAsciiTable(YamlTestBase):
    r"""Test connection with AsciiTable."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input: {{ %s }}' % _yaml_env,
                  '    output: inputA',
                  '    read_meth: table',
                  '  - input: outputA',
                  '    output: output.txt',
                  '    write_meth: table'],)


class TestYamlConnectionInputAsciiTableArray(YamlTestBase):
    r"""Test connection with AsciiTable as array."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input: {{ %s }}' % _yaml_env,
                  '    output: inputA',
                  '    read_meth: table_array',
                  '    field_units: name,count,size',
                  '    format_str: "%5s\t%d\t%f\n"',
                  '  - input: outputA',
                  '    output: output.txt',
                  '    write_meth: table_array',
                  '    field_units: name,count,size',
                  '    format_str: "%5s\t%d\t%f\n"'], )


class TestYamlConnectionInputPickle(YamlTestBase):
    r"""Test connection with Pickle."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input: {{ %s }}' % _yaml_env,
                  '    output: inputA',
                  '    read_meth: pickle',
                  '  - input: outputA',
                  '    output: output.txt',
                  '    write_meth: pickle'], )


class TestYamlConnectionInputPandas(YamlTestBase):
    r"""Test connection with Pandas csv."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input: {{ %s }}' % _yaml_env,
                  '    output: inputA',
                  '    read_meth: pandas',
                  '    field_units: name,count,size',
                  '    format_str: "%5s\t%d\t%f\n"',
                  '  - input: outputA',
                  '    output: output.txt',
                  '    write_meth: pandas',
                  '    field_units: name,count,size',
                  '    format_str: "%5s\t%d\t%f\n"'], )


class TestYamlConnectionInputAsciiMap(YamlTestBase):
    r"""Test connection with AsciiMap."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input: {{ %s }}' % _yaml_env,
                  '    output: inputA',
                  '    read_meth: map',
                  '  - input: outputA',
                  '    output: output.txt',
                  '    write_meth: map'], )


class TestYamlConnectionInputPly(YamlTestBase):
    r"""Test connection with Ply file."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input: {{ %s }}' % _yaml_env,
                  '    output: inputA',
                  '    read_meth: ply',
                  '  - input: outputA',
                  '    output: output.ply',
                  '    write_meth: ply'], )


class TestYamlConnectionInputObj(YamlTestBase):
    r"""Test connection with Obj file."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input: {{ %s }}' % _yaml_env,
                  '    output: inputA',
                  '    read_meth: obj',
                  '  - input: outputA',
                  '    output: output.obj',
                  '    write_meth: obj'], )


class TestYamlComponentError(YamlTestBaseError):
    r"""Test error for non-dictionary component."""
    _error = ValidationError
    _contents = (['models: error'],)


class TestYamlDuplicateError(YamlTestBaseError):
    r"""Test error when there are two components with the same name."""
    _error = ValueError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c'],)


class TestYamlConnectionError(YamlTestBaseError):
    r"""Test error when there is not connection for a model I/O channel."""
    _error = RuntimeError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA'],)


class TestYamlConnectionError_forkin(YamlTestBaseError):
    r"""Test error when there is not connection for a fork input channel."""
    _error = RuntimeError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  'connections:',
                  '  - input:',
                  '      - output1',
                  '      - output2',
                  '    output: inputA'],)


# Error not raised as both outputs could be files
# class TestYamlConnectionError_forkout(YamlTestBaseError):
#     r"""Test error when there is not connection for a fork output channel."""
#     _error = RuntimeError
#     _contents = (['models:',
#                   '  - name: modelA',
#                   '    driver: GCCModelDriver',
#                   '    args: ./src/modelA.c',
#                   '    outputs:',
#                   '      - outputA',
#                   'connections:',
#                   '  - input: outputA',
#                   '    output:',
#                   '      - input1',
#                   '      - input2'],)


class TestYamlConnectionError_readmeth(YamlTestBaseError):
    r"""Test error when read_meth is specified for non-file."""
    _error = ValidationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '',
                  'connections:',
                  '  - input: outputB',
                  '    output: inputA',
                  '    read_meth: pickle'],
                 ['models:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c',
                  '    outputs:',
                  '      - outputB'],)


class TestYamlConnectionError_writemeth(YamlTestBaseError):
    r"""Test error when write_meth is specified for non-file."""
    _error = ValidationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '',
                  'connections:',
                  '  - input: outputB',
                  '    output: inputA',
                  '    write_meth: pickle'],
                 ['models:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c',
                  '    outputs:',
                  '      - outputB'],)


class TestYamlMissingModelArgsError(YamlTestBaseError):
    r"""Test error when there is a missing arguments to a model."""
    _error = ValidationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    inputs:',
                  '      name: inputA',
                  '      driver: FileInputDriver',
                  '      args: {{ %s }}' % _yaml_env],)


class TestYamlMissingIOArgsError_input(YamlTestBaseError):
    r"""Test error when there is a missing arguments to an input driver."""
    _error = ValidationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      name: inputA',
                  '      driver: FileInputDriver'],)


class TestYamlMissingIOArgsError_output(YamlTestBaseError):
    r"""Test error when there is a missing arguments to an output driver."""
    _error = ValidationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    outputs:',
                  '      name: outputA',
                  '      driver: FileOutputDriver'],)


class TestYamlMissingConnArgsError(YamlTestBaseError):
    r"""Test error when there is a missing arguments to a connection."""
    _error = ValidationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '',
                  'connections:',
                  '  - input: outputB'],
                 ['models:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c',
                  '    outputs:',
                  '      - outputB'],)


class TestYamlMissingConnInputError(YamlTestBaseError):
    r"""Test error when there is no model output matching connection input."""
    _error = RuntimeError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '',
                  'connections:',
                  '  - input: outputB',
                  '    output: inputA'],
                 ['models:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c'],)


class TestYamlMissingConnInputFileError(YamlTestBaseError):
    r"""Test error when there is no file for missing connection input."""
    _error = ValidationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '',
                  'connections:',
                  '  - input: outputB',
                  '    output: inputA'],
                 ['models:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c'],)


class TestYamlMissingConnIOError(YamlTestBaseError):
    r"""Test error when there is no model input/output matching connection."""
    _error = ValidationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '',
                  'connections:',
                  '  - input: {{ %s }}' % _yaml_env,
                  '    output: inputA'],)


class TestYamlConnectionInputFileReadMethError(YamlTestBaseError):
    r"""Test error for invalid read_meth."""
    _error = ValidationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input: {{ %s }}' % _yaml_env,
                  '    output: inputA',
                  '    read_meth: invalid',
                  '  - input: outputA',
                  '    output: output.txt'],)


class TestYamlConnectionInputFileWriteMethError(YamlTestBaseError):
    r"""Test error for invalid write_meth."""
    _error = ValidationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input: {{ %s }}' % _yaml_env,
                  '    output: inputA',
                  '  - input: outputA',
                  '    output: output.txt',
                  '    write_meth: invalid'],)



import os
import uuid
from yggdrasil import backwards
from yggdrasil.tests import YggTestClassInfo, assert_equal
from yggdrasil.communication import new_comm, get_comm, CommBase


def test_registry():
    r"""Test registry of comm."""
    comm_class = 'CommBase'
    key = 'key1'
    value = None
    assert(not CommBase.is_registered(comm_class, key))
    assert(not CommBase.unregister_comm(comm_class, key))
    assert_equal(CommBase.get_comm_registry(None), {})
    assert_equal(CommBase.get_comm_registry(comm_class), {})
    CommBase.register_comm(comm_class, key, value)
    assert(key in CommBase.get_comm_registry(comm_class))
    assert(CommBase.is_registered(comm_class, key))
    assert(not CommBase.unregister_comm(comm_class, key, dont_close=True))
    CommBase.register_comm(comm_class, key, value)
    assert(not CommBase.unregister_comm(comm_class, key))


class TestCommBase(YggTestClassInfo):
    r"""Tests for CommBase communication class.

    Attributes:
        send_inst_kwargs (dict): Keyword arguments for send half of the comm
            pair.

    """

    comm = 'CommBase'
    attr_list = ['name', 'address', 'direction',
                 'serializer', 'recv_timeout',
                 'close_on_eof_recv', 'opp_address', 'opp_comms',
                 'maxMsgSize']
    
    @property
    def cleanup_comm_classes(self):
        r"""list: Comm classes that should be cleaned up following the test."""
        return set([self.comm, self.send_inst_kwargs['comm']])

    @property
    def name(self):
        r"""str: Name of the test connection."""
        return 'Test%s_%s' % (self.cls, self.uuid)

    @property
    def cls(self):
        r"""str: Communication class."""
        return self.comm

    @property
    def mod(self):
        r"""str: Absolute module import."""
        return 'yggdrasil.communication.%s' % self.cls

    @property
    def is_installed(self):
        r"""bool: Is the communication class installed."""
        return self.import_cls.is_installed(language='python')

    @property
    def send_inst_kwargs(self):
        r"""dict: Keyword arguments for send instance."""
        out = {'comm': self.comm, 'reverse_names': True, 'direction': 'send'}
        out.update(self.testing_options['kwargs'])
        return out

    @property
    def inst_args(self):
        r"""list: Arguments for tested class."""
        return [self.name]

    @property
    def inst_kwargs(self):
        r"""dict: Keyword arguments for tested class."""
        return self.send_instance.opp_comm_kwargs()

    @property
    def recv_instance(self):
        r"""Alias for instance."""
        return self.instance

    @property
    def maxMsgSize(self):
        r"""int: Maximum message size."""
        return self.instance.maxMsgSize

    @property
    def test_msg_array(self):
        r"""str: Test message that should be used for any send/recv tests."""
        return self.testing_options.get('msg_array', None)
    
    @property
    def test_msg(self):
        r"""str: Test message that should be used for any send/recv tests."""
        return self.testing_options['msg']

    @property
    def msg_long(self):
        r"""str: Small test message for sending."""
        out = self.test_msg
        if isinstance(out, backwards.bytes_type):
            out += (self.maxMsgSize * b'0')
        return out
            
    def setup(self, *args, **kwargs):
        r"""Initialize comm object pair."""
        assert(self.is_installed)
        sleep_after_connect = kwargs.pop('sleep_after_connect', False)
        send_inst_kwargs = self.send_inst_kwargs
        kwargs.setdefault('nprev_comm', self.comm_count)
        kwargs.setdefault('nprev_fd', self.fd_count)
        self.send_instance = new_comm(self.name, **send_inst_kwargs)
        super(TestCommBase, self).setup(*args, **kwargs)
        if sleep_after_connect:
            self.send_instance.sleep()
        # CommBase is dummy class that never opens
        if self.comm in ['CommBase', 'AsyncComm']:
            assert(not self.send_instance.is_open)
            assert(not self.recv_instance.is_open)
        else:
            assert(self.send_instance.is_open)
            assert(self.recv_instance.is_open)

    def teardown(self, *args, **kwargs):
        r"""Destroy comm object pair."""
        self.remove_instance(self.send_instance)
        super(TestCommBase, self).teardown(*args, **kwargs)

    def create_instance(self):
        r"""Create a new instance of the class."""
        inst = get_comm(*self.inst_args, **self.inst_kwargs)
        assert(isinstance(inst, self.import_cls))
        return inst

    def remove_instance(self, inst):
        r"""Remove an instance."""
        inst.close()
        assert(inst.is_closed)
        super(TestCommBase, self).remove_instance(inst)

    def get_fresh_error_instance(self, recv=False):
        r"""Get comm instance with ErrorClass parent class."""
        try:
            send_kwargs = self.send_inst_kwargs
            err_kwargs = dict(base_comm=send_kwargs['comm'], new_comm_class='ErrorComm')
            err_name = self.name + '_' + self.uuid
            if not recv:
                send_kwargs.update(**err_kwargs)
            send_inst = new_comm(err_name, **send_kwargs)
            recv_kwargs = send_inst.opp_comm_kwargs()
            recv_kwargs['comm'] = send_kwargs['comm']
            if recv:
                recv_kwargs.update(**err_kwargs)
            recv_inst = new_comm(err_name, **recv_kwargs)
        except BaseException as e:  # pragma: debug
            print(e)
            raise
        return send_inst, recv_inst

    def test_empty_obj_recv(self):
        r"""Test identification of empty message."""
        msg = self.instance.empty_obj_recv
        assert(self.instance.is_empty_recv(msg))
        assert(not self.instance.is_empty_recv(self.instance.eof_msg))
        if self.recv_instance.recv_converter is None:
            self.recv_instance.recv_converter = lambda x: x
            msg = self.instance.empty_obj_recv
            assert(self.instance.is_empty_recv(msg))
            assert(not self.instance.is_empty_recv(self.instance.eof_msg))
            
    def test_error_name(self):
        r"""Test error on missing address."""
        self.assert_raises(RuntimeError, self.import_cls, 'test%s' % uuid.uuid4())

    def test_error_send(self):
        r"""Test error on send."""
        send_inst, recv_inst = self.get_fresh_error_instance()
        send_inst._first_send_done = True
        send_inst.error_replace('send_multipart')
        flag = send_inst.send(self.test_msg)
        assert(not flag)
        send_inst.restore_all()
        send_inst.close()
        recv_inst.close()

    def test_error_recv(self):
        r"""Test error on recv."""
        self.fd_count
        send_inst, recv_inst = self.get_fresh_error_instance(recv=True)
        self.fd_count
        recv_inst.error_replace('recv_multipart')
        flag, msg_recv = recv_inst.recv()
        self.fd_count
        assert(not flag)
        recv_inst.restore_all()
        send_inst.close()
        recv_inst.close()
        self.fd_count

    def test_send_recv_after_close(self):
        r"""Test that opening twice dosn't cause errors and that send/recv after
        close returns false."""
        self.send_instance.open()
        self.recv_instance.open()
        if self.comm in ['RMQComm', 'RMQAsyncComm']:
            self.send_instance.bind()
            self.recv_instance.bind()
        self.send_instance.close()
        self.recv_instance.close()
        assert(self.send_instance.is_closed)
        assert(self.recv_instance.is_closed)
        flag = self.send_instance.send(self.test_msg)
        assert(not flag)
        flag, msg_recv = self.recv_instance.recv()
        assert(not flag)

    def test_attributes(self):
        r"""Assert that the instance has all of the required attributes."""
        for a in self.attr_list:
            if not hasattr(self.send_instance, a):  # pragma: debug
                raise AttributeError("Send comm does not have attribute %s" % a)
            if not hasattr(self.recv_instance, a):  # pragma: debug
                raise AttributeError("Recv comm does not have attribute %s" % a)
            getattr(self.send_instance, a)
            getattr(self.recv_instance, a)
        self.instance.debug('maxMsgSize: %d, %d, %d', self.maxMsgSize,
                            self.send_instance.maxMsgSize,
                            self.recv_instance.maxMsgSize)
        self.instance.opp_comm_kwargs()
        if self.import_cls.is_file:
            assert(self.import_cls.is_installed(language='invalid'))
        else:
            assert(not self.import_cls.is_installed(language='invalid'))

    def test_invalid_direction(self):
        r"""Check that error raised for invalid direction."""
        kwargs = self.send_inst_kwargs
        kwargs['direction'] = 'invalid'
        self.assert_raises(ValueError, new_comm, self.name + "_" + self.uuid,
                           **kwargs)

    def test_work_comm(self):
        r"""Test creating/removing a work comm."""
        wc_send = self.instance.create_work_comm()
        self.assert_raises(KeyError, self.instance.add_work_comm, wc_send)
        # Create recv instance in way that tests new_comm
        header_recv = dict(id=self.uuid + '1', address=wc_send.address)
        recv_kwargs = self.instance.get_work_comm_kwargs
        recv_kwargs['work_comm_name'] = 'test_worker_%s' % header_recv['id']
        recv_kwargs['new_comm_class'] = wc_send.comm_class
        os.environ[recv_kwargs['work_comm_name']] = wc_send.opp_address
        wc_recv = self.instance.create_work_comm(**recv_kwargs)
        # wc_recv = self.instance.get_work_comm(header_recv)
        if self.comm in ['CommBase', 'AsyncComm']:
            flag = wc_send.send(self.test_msg)
            assert(not flag)
            flag, msg_recv = wc_recv.recv()
            assert(not flag)
        else:
            flag = wc_send.send(self.test_msg)
            assert(flag)
            flag, msg_recv = wc_recv.recv(self.timeout)
            assert(flag)
            self.assert_equal(msg_recv, self.test_msg)
            # Assert errors on second attempt
            # self.assert_raises(RuntimeError, wc_send.send, self.test_msg)
            self.assert_raises(RuntimeError, wc_recv.recv)
        self.instance.remove_work_comm(wc_send.uuid)
        self.instance.remove_work_comm(wc_recv.uuid)
        self.instance.remove_work_comm(wc_recv.uuid)
        # Create work comm that should be cleaned up on teardown
        self.instance.create_work_comm()

    def map_sent2recv(self, obj):
        r"""Convert a sent object into a received one."""
        return obj

    def assert_msg_equal(self, x, y):
        r"""Assert that two messages are equivalent."""
        if not (isinstance(y, type(self.send_instance.eof_msg))
                and (y == self.send_instance.eof_msg)):
            y = self.map_sent2recv(y)
        self.assert_equal(x, y)

    def assert_msg_lists_equal(self, x, y):
        r"""Assert that two lists of messages are equivalent."""
        self.assert_equal(len(x), len(y))
        for ix, iy in zip(x, y):
            self.assert_msg_equal(ix, iy)

    def recv_message_list(self, recv_inst, expected_result=None,
                          break_on_empty=False):
        r"""Continue receiving from a receive instance until flag is False (or
        an empty messages is received and break_on_empty is True). On receipt of
        a False flag, the recieved message is checked against the EOF message.

        Args:
            recv_inst (yggdrasil.communication.CommBase.CommBase): Communication
                instance that should be received from.
            expected_result (list, optional): A list of messages that the
                recieved messages should be compared against. Defaults to None
                and is ignored.
            break_on_empty (bool, optional): If True, messages will stop being
                received from the communication instance when an empty message
                is received. Defaults to False.

        Returns:
            list: Received messages.

        """
        flag = True
        msg_list = []
        while flag:
            flag, msg_recv = recv_inst.recv()
            if flag:
                if break_on_empty and recv_inst.is_empty_recv(msg_recv):
                    break
                msg_list.append(msg_recv)
            else:
                self.assert_equal(msg_recv, recv_inst.eof_msg)
        if expected_result is not None:
            self.assert_msg_lists_equal(msg_list, expected_result)
        return msg_list

    def do_send_recv(self, send_meth='send', recv_meth='recv',
                     msg_send=None, msg_recv=None,
                     n_msg_send_meth='n_msg_send', n_msg_recv_meth='n_msg_recv',
                     reverse_comms=False, send_kwargs=None, recv_kwargs=None,
                     n_send=1, n_recv=1, print_status=False,
                     close_on_send_eof=None, close_on_recv_eof=None):
        r"""Generic send/recv of a message."""
        tkey = 'do_send_recv'
        is_eof = ('eof' in send_meth)
        if msg_send is None:
            if is_eof:
                msg_send = self.send_instance.eof_msg
            else:
                msg_send = self.test_msg
        if msg_recv is None:
            msg_recv = msg_send
        # print('send', msg_send)
        # print('recv', msg_recv)
        if send_kwargs is None:
            send_kwargs = dict()
        if recv_kwargs is None:
            recv_kwargs = dict()
        if is_eof:
            send_args = tuple()
        else:
            send_args = (msg_send,)
        self.assert_equal(getattr(self.send_instance, n_msg_send_meth), 0)
        self.assert_equal(getattr(self.recv_instance, n_msg_recv_meth), 0)
        if reverse_comms:
            send_instance = self.recv_instance
            recv_instance = self.send_instance
        else:
            send_instance = self.send_instance
            recv_instance = self.recv_instance
        if close_on_recv_eof is None:
            close_on_recv_eof = recv_instance.close_on_eof_recv
        if close_on_send_eof is None:
            close_on_send_eof = send_instance.close_on_eof_send
        recv_instance.close_on_eof_recv = close_on_recv_eof
        send_instance.close_on_eof_send = close_on_send_eof
        if self.comm == 'ForkComm':
            for x in recv_instance.comm_list:
                x.close_on_eof_recv = close_on_recv_eof
            for x in send_instance.comm_list:
                x.close_on_eof_send = close_on_send_eof
        fsend_meth = getattr(send_instance, send_meth)
        frecv_meth = getattr(recv_instance, recv_meth)
        if self.comm in ['CommBase', 'AsyncComm']:
            flag = fsend_meth(*send_args, **send_kwargs)
            assert(not flag)
            flag, msg_recv0 = frecv_meth(**recv_kwargs)
            assert(not flag)
            if self.comm == 'CommBase':
                self.assert_raises(NotImplementedError, self.recv_instance._send,
                                   self.test_msg)
                self.assert_raises(NotImplementedError, self.recv_instance._recv)
        else:
            for i in range(n_send):
                flag = fsend_meth(*send_args, **send_kwargs)
                assert(flag)
            # Wait for messages to be received
            for i in range(n_recv):
                if not is_eof:
                    T = recv_instance.start_timeout(self.timeout, key_suffix=tkey)
                    while ((not T.is_out) and (not recv_instance.is_closed)
                           and (getattr(recv_instance,
                                        n_msg_recv_meth) == 0)):  # pragma: debug
                        recv_instance.sleep()
                    recv_instance.stop_timeout(key_suffix=tkey)
                    assert(getattr(recv_instance, n_msg_recv_meth) >= 1)
                    # IPC nolimit sends multiple messages
                    # self.assert_equal(recv_instance.n_msg_recv, 1)
                flag, msg_recv0 = frecv_meth(timeout=self.timeout, **recv_kwargs)
                if is_eof and close_on_recv_eof:
                    assert(not flag)
                    assert(recv_instance.is_closed)
                else:
                    assert(flag)
                self.assert_msg_equal(msg_recv0, msg_recv)
            # Wait for send to close
            if is_eof and close_on_send_eof:
                T = send_instance.start_timeout(self.timeout, key_suffix=tkey)
                while (not T.is_out) and (not send_instance.is_closed):  # pragma: debug
                    send_instance.sleep()
                send_instance.stop_timeout(key_suffix=tkey)
                assert(send_instance.is_closed)
        # Make sure no messages outgoing
        T = send_instance.start_timeout(self.timeout, key_suffix=tkey)
        while ((not T.is_out) and (getattr(send_instance,
                                           n_msg_send_meth) != 0)):  # pragma: debug
            send_instance.sleep()
        send_instance.stop_timeout(key_suffix=tkey)
        # Print status of comms
        if print_status:
            send_instance.printStatus()
            recv_instance.printStatus()
        # else:
        #     send_instance.get_status_message()
        #     recv_instance.get_status_message()
        # Confirm recept of messages
        if not (is_eof or reverse_comms):
            send_instance.wait_for_confirm(timeout=self.timeout)
            recv_instance.wait_for_confirm(timeout=self.timeout)
            assert(send_instance.is_confirmed)
            assert(recv_instance.is_confirmed)
            send_instance.confirm(noblock=True)
            recv_instance.confirm(noblock=True)
        self.assert_equal(getattr(send_instance, n_msg_send_meth), 0)
        self.assert_equal(getattr(recv_instance, n_msg_recv_meth), 0)

    def test_drain_messages(self):
        r"""Test waiting for messages to drain."""
        self.send_instance.drain_messages(timeout=self.timeout)
        self.assert_equal(self.send_instance.n_msg_send_drain, 0)
        if not self.recv_instance.is_file:
            self.recv_instance.drain_messages(timeout=self.timeout)
            self.assert_equal(self.recv_instance.n_msg_recv_drain, 0)
        self.assert_raises(ValueError, self.send_instance.drain_messages,
                           variable='n_msg_invalid')
        self.assert_raises(ValueError, self.recv_instance.drain_messages,
                           variable='n_msg_invalid')

    def test_recv_nomsg(self):
        r"""Test recieve when there is no waiting message."""
        flag, msg_recv = self.recv_instance.recv(timeout=self.sleeptime)
        if self.comm in ['CommBase', 'AsyncComm']:
            assert(not flag)
        else:
            assert(flag)
        assert(not msg_recv)

    def test_send_recv(self):
        r"""Test send/recv of a small message."""
        self.do_send_recv(print_status=True)

    def test_send_recv_nolimit(self):
        r"""Test send/recv of a large message."""
        assert(len(self.msg_long) > self.maxMsgSize)
        self.do_send_recv('send_nolimit', 'recv_nolimit', self.msg_long,
                          print_status=True)

    def test_send_recv_array(self):
        r"""Test send/recv of a array message."""
        msg_send = getattr(self, 'test_msg_array', None)
        self.do_send_recv('send_array', 'recv_array', msg_send=msg_send)

    def test_eof(self):
        r"""Test send/recv of EOF message."""
        self.do_send_recv(send_meth='send_eof')

    def test_eof_no_close(self):
        r"""Test send/recv of EOF message with no close."""
        self.do_send_recv(send_meth='send_eof', close_on_recv_eof=False)

    def test_purge(self, nrecv=1):
        r"""Test purging messages from the comm."""
        self.assert_equal(self.send_instance.n_msg, 0)
        self.assert_equal(self.recv_instance.n_msg, 0)
        # Purge recv while open
        if self.comm not in ['CommBase', 'AsyncComm']:
            flag = self.send_instance.send(self.test_msg)
            assert(flag)
            T = self.recv_instance.start_timeout()
            while ((not T.is_out) and (self.recv_instance.n_msg
                                       != nrecv)):  # pragma: debug
                self.recv_instance.sleep()
            self.recv_instance.stop_timeout()
            self.assert_greater(self.recv_instance.n_msg, 0)
        self.recv_instance.purge()
        # Uni-directional comms can't know about messages sent
        # self.assert_equal(self.send_instance.n_msg, 0)
        self.assert_equal(self.recv_instance.n_msg, 0)
        # Purge recv while closed
        self.recv_instance.close()
        self.recv_instance.purge()

    def test_send_recv_dict(self):
        r"""Test send/recv message as dict."""
        msg_send = self.testing_options['dict']
        self.do_send_recv(send_meth='send_dict', recv_meth='recv_dict',
                          msg_send=msg_send)
        
    def test_send_recv_dict_names(self):
        r"""Test send/recv message as dict with names."""
        msg_send = self.testing_options['dict']
        field_order = self.testing_options.get('field_names', None)
        if field_order is not None:
            self.do_send_recv(send_meth='send_dict', recv_meth='recv_dict',
                              msg_send=msg_send,
                              send_kwargs={'field_order': field_order},
                              recv_kwargs={'field_order': field_order})



import os
from yggdrasil.config import ygg_cfg
from yggdrasil.tests import assert_equal, assert_raises, YggTestClass
from yggdrasil.drivers import CompiledModelDriver
import yggdrasil.drivers.tests.test_ModelDriver as parent


def test_get_compilation_tool_registry():
    r"""Test errors raised by get_compilation_tool_registry."""
    assert_raises(ValueError, CompiledModelDriver.get_compilation_tool_registry,
                  'invalid')

    
def test_find_compilation_tool():
    r"""Test errors raised by find_compilation_tool."""
    assert_raises(RuntimeError, CompiledModelDriver.find_compilation_tool,
                  'archiver', 'cmake')
    

def test_get_compilation_tool():
    r"""Test get_compilation_tool for different name variations."""
    from yggdrasil.drivers.CModelDriver import CModelDriver
    tooltype = 'compiler'
    out = CModelDriver.get_tool('compiler').__class__
    toolname = out.toolname.lower()
    toolpath = os.path.join('somedir', toolname)
    toolfile = toolpath + '.exe'
    vals = [toolname.upper(), toolpath, toolfile, toolfile.upper()]
    for v in vals:
        assert_equal(CompiledModelDriver.get_compilation_tool(tooltype, v), out)
    assert_raises(ValueError, CompiledModelDriver.get_compilation_tool,
                  'compiler', 'invalid')


def test_CompilationToolBase():
    r"""Test error in CompilationToolBase."""
    assert_raises(RuntimeError, CompiledModelDriver.CompilationToolBase,
                  invalid='invalid')


class DummyCompiler(CompiledModelDriver.CompilerBase):
    r"""Dummy test class."""
    _dont_register = True
    toolname = 'dummy'
    languages = ['dummy']
    search_path_env = 'PATH'
    _language_ext = ['.c']
    default_linker = False
    default_archiver = None
    combine_with_linker = True
    compile_only_flag = None


class TestCompilationTool(YggTestClass):
    r"""Test class for compilation tools."""

    _mod = 'yggdrasil.drivers.CompiledModelDriver'
    _cls = 'CompilationToolBase'

    def test_append_flags(self):
        r"""Test append_flags."""
        self.assert_raises(ValueError, self.import_cls.append_flags,
                           [], '-T', 'bye', invalid='invalid')
        self.assert_raises(ValueError, self.import_cls.append_flags,
                           ['-Thello'], '-T%s', 'bye', no_duplicates=True)
        kws_list = [((['a', 'b', 'c'], '0', '1'), {'prepend': True},
                     ['0', '1', 'a', 'b', 'c']),
                    ((['a', 'b', 'c'], '0', '1'), {'position': -1},
                     ['a', 'b', 'c', '0', '1']),
                    ((['a', 'b', 'c'], '0', '1'), {'position': -2},
                     ['a', 'b', '0', '1', 'c'])]
        for (out, key, val), kws, res in kws_list:
            self.import_cls.append_flags(out, key, val, **kws)
            self.assert_equal(out, res)

    def test_create_flag(self):
        r"""Test create_flag."""
        test_args = [({'key': '-T%s'}, ['a', 'b', 'c'], ['-Ta', '-Tb', '-Tc']),
                     ('-set-this', True, ['-set-this']),
                     ('-set-this', False, []),
                     ('-set-this', None, [])]
        for (key, val, out) in test_args:
            self.assert_equal(self.import_cls.create_flag(key, val), out)

    def test_not_implemented(self):
        r"""Test raising of NotImplementedErrors for incomplete classes."""
        pass

    def test_get_flags(self):
        r"""Test get_flags."""
        self.assert_equal(self.import_cls.get_flags(flags='hello'), ['hello'])

    def test_get_search_path(self):
        r"""Test get_search_path."""
        if self._cls == 'CompilationToolBase':
            self.assert_raises(NotImplementedError, self.import_cls.get_search_path)
        else:
            self.import_cls.get_search_path()
            
    def test_get_executable_command(self):
        r"""Test get_executable_command."""
        if self.import_cls.toolname is None:
            self.assert_raises(NotImplementedError,
                               self.import_cls.get_executable_command, [])
            

class TestDummyCompiler(TestCompilationTool):
    r"""Test class for DummyCompiler."""

    _mod = 'yggdrasil.drivers.tests.test_CompiledModelDriver'
    _cls = 'DummyCompiler'

    def __init__(self, *args, **kwargs):
        super(TestDummyCompiler, self).__init__(*args, **kwargs)
        self._inst_kwargs['linker'] = False

    def test_call(self):
        r"""Test call."""
        self.assert_raises(RuntimeError, self.import_cls.call, 'args',
                           out='test', dont_link=True)

    def test_linker(self):
        r"""Test linker."""
        self.assert_equal(self.import_cls.linker(), False)

    def test_archiver(self):
        r"""Test archiver."""
        self.assert_raises(RuntimeError, self.import_cls.archiver)
        
    def test_get_flags(self):
        r"""Test get_flags."""
        self.assert_equal(self.import_cls.get_flags(flags='hello',
                                                    libtype='object'), ['hello'])
        
    def test_get_executable_command(self):
        r"""Test get_executable_command."""
        self.import_cls.get_executable_command([], dont_link=True)
        

class TestCompiledModelParam(parent.TestModelParam):
    r"""Test parameters for basic CompiledModelDriver class."""

    driver = 'CompiledModelDriver'
    
    def __init__(self, *args, **kwargs):
        super(TestCompiledModelParam, self).__init__(*args, **kwargs)
        self.attr_list += ['source_files']
        for k in ['compiler', 'linker', 'archiver']:
            self.attr_list += [k, '%s_flags' % k, '%s_tool' % k]
        if (self.src is not None) and self.import_cls.is_installed():
            self.args = [self.import_cls.get_tool('compiler').get_output_file(
                self.src[0])]
            self._inst_kwargs['source_files'] = self.src


class TestCompiledModelDriverNoInit(TestCompiledModelParam,
                                    parent.TestModelDriverNoInit):
    r"""Test runner for CompiledModelDriver without creating an instance."""
    
    def test_build(self):
        r"""Test building libraries as a shared/static library or object files."""
        self.import_cls.compile_dependencies(libtype='shared', overwrite=True)
        self.import_cls.compile_dependencies(libtype='object', overwrite=True)
        self.import_cls.compile_dependencies(libtype='static', overwrite=True)
        self.import_cls.compile_dependencies(libtype='shared', overwrite=False)

    def test_get_tool(self):
        r"""Test other methods of calling get_tool."""
        self.import_cls.get_tool('compiler', return_prop='name')
        self.import_cls.get_tool('compiler', return_prop='flags')
        self.assert_raises(ValueError, self.import_cls.get_tool, 'compiler',
                           return_prop='invalid')

    def test_get_dependency_source(self):
        r"""Test get_dependency_source."""
        dep_list = (list(self.import_cls.external_libraries.keys())
                    + list(self.import_cls.internal_libraries.keys()))
        for dep in dep_list:
            self.import_cls.get_dependency_source(dep, default='default')
        self.assert_raises(ValueError, self.import_cls.get_dependency_source,
                           'invalid')
        self.assert_equal(self.import_cls.get_dependency_source(__file__),
                          __file__)
        self.assert_equal(self.import_cls.get_dependency_source(
            'invalid', default='default'), 'default')

    def test_get_dependency_library(self):
        r"""Test get_dependency_library."""
        self.assert_raises(ValueError, self.import_cls.get_dependency_library,
                           'invalid', libtype='invalid')
        for dep, info in self.import_cls.external_libraries.items():
            libtype_orig = info.get('libtype', None)
            if libtype_orig not in ['static', 'shared']:
                continue
            if libtype_orig == 'static':  # pragma: no cover
                libtype = 'shared'
            else:
                libtype = 'static'
            self.assert_raises(ValueError, self.import_cls.get_dependency_library,
                               dep, libtype=libtype)
        self.assert_raises(ValueError, self.import_cls.get_dependency_library,
                           'invalid')
        self.assert_equal(self.import_cls.get_dependency_library(__file__),
                          __file__)
        self.assert_equal(self.import_cls.get_dependency_library(
            'invalid', default='default'), 'default')

    def test_get_dependency_include_dirs(self):
        r"""Test get_dependency_include_dirs."""
        self.assert_raises(ValueError, self.import_cls.get_dependency_include_dirs,
                           'invalid')
        self.assert_equal(self.import_cls.get_dependency_include_dirs(__file__),
                          [os.path.dirname(__file__)])
        self.assert_equal(self.import_cls.get_dependency_include_dirs(
            'invalid', default='default'), ['default'])

    def test_get_dependency_order(self):
        r"""Test get_dependency_order."""
        deps = list(self.import_cls.internal_libraries.keys())
        self.import_cls.get_dependency_order(deps)

    def test_get_flags(self):
        r"""Test get_flags."""
        compiler = self.import_cls.get_tool('compiler')
        if compiler:
            if ((compiler.combine_with_linker
                 or compiler.no_separate_linking)):
                print(compiler, compiler.get_flags(invalid_kw=True,
                                                   unused_kwargs={},
                                                   libraries=[]))
            else:
                self.assert_raises(ValueError, compiler.get_flags)

    def test_get_linker_flags(self):
        r"""Test get_linker_flags."""
        if self.import_cls.get_tool('archiver') is False:
            self.assert_raises(RuntimeError, self.import_cls.get_linker_flags,
                               libtype='static')
        else:
            self.import_cls.get_linker_flags(libtype='static', for_model=True,
                                             use_library_path_internal=True)
        if getattr(self.import_cls.get_tool('linker'), 'is_dummy', False):
            self.assert_raises(RuntimeError, self.import_cls.get_linker_flags,
                               libtype='shared')
        else:
            self.import_cls.get_linker_flags(libtype='shared', for_model=True,
                                             use_library_path=True)
            self.import_cls.get_linker_flags(libtype='shared', for_model=True,
                                             skip_library_libs=True,
                                             use_library_path=True)
            self.import_cls.get_linker_flags(libtype='shared', for_model=True,
                                             skip_library_libs=True,
                                             use_library_path_internal=True)

    def test_executable_command(self):
        r"""Test executable_command."""
        self.assert_raises(ValueError, self.import_cls.executable_command, ['test'],
                           exec_type='invalid')
        if self.import_cls.get_tool('compiler').no_separate_linking:
            self.assert_raises(RuntimeError, self.import_cls.executable_command,
                               ['test'], exec_type='linker')
        else:
            self.import_cls.executable_command(['test'], exec_type='linker')

    def test_compiler_call(self):
        r"""Test compiler call."""
        tool = self.import_cls.get_tool('compiler')
        self.assert_equal(tool.call('args', out='test',
                                    dry_run=True, skip_flags=True), '')
        src = [x + tool.get_language_ext()[0] for x in ['args1', 'args2']]
        self.assert_raises(ValueError, tool.call, src,
                           out='out1', dont_link=True)
        kwargs = dict(dry_run=True, working_dir=os.getcwd())
        if self.import_cls.language in ['cmake']:
            src = src[:1]
            kwargs['target'] = tool.file2base(src[0])
        tool.call(src, dont_link=True, **kwargs)
        tool.call(src, **kwargs)

    def test_configure(self):
        r"""Test configuration (presumably after it has already been done)."""
        self.import_cls.configure(ygg_cfg)
        
    def test_get_output_file(self):
        r"""Test get_output_file."""
        tool = self.import_cls.get_tool('compiler')
        fname = self.src[0]
        tool.get_output_file([fname])

    def test_invalid_function_param(self):
        r"""Test errors raise during class creation when parameters are invalid."""
        super(TestCompiledModelDriverNoInit, self).test_invalid_function_param()
        kwargs = dict(self.inst_kwargs)
        kwargs['name'] = 'test'
        kwargs['args'] = ['test']
        kwargs['function'] = 'invalid'
        kwargs['source_files'] = []
        # With source_files
        if self.import_cls.language_ext:
            kwargs['source_files'] = ['invalid' + self.import_cls.language_ext[0]]
        self.assert_raises(ValueError, self.import_cls, **kwargs)
        

class TestCompiledModelDriverNoStart(TestCompiledModelParam,
                                     parent.TestModelDriverNoStart):
    r"""Test runner for CompiledModelDriver without start."""

    def test_compilers(self):
        r"""Test available compilers."""
        # Record old tools
        old_tools = {}
        for k in ['compiler', 'linker', 'achiver']:
            old_tools[k] = getattr(self.instance, '%s_tool' % k, None)
        # Compile with each compiler
        for k, v in self.import_cls.get_available_tools('compiler').items():
            if not v.is_installed():
                continue
            setattr(self.instance, 'compiler_tool', v)
            setattr(self.instance, 'linker_tool', v.linker())
            setattr(self.instance, 'archiver_tool', v.archiver())
            self.instance.compile_model()
        # Restore the old tools
        for k, v in old_tools.items():
            setattr(self.instance, '%s_tool' % k, v)

    def test_compile_model(self):
        r"""Test compile model with alternate set of input arguments."""
        fname = self.src[0]
        self.assert_raises(RuntimeError, self.instance.compile_model,
                           out=os.path.basename(fname),
                           working_dir=os.path.dirname(fname),
                           overwrite=True)
        
    # Done in driver, but driver not started
    def teardown(self):
        r"""Remove the instance, stoppping it."""
        self.instance.cleanup()
        super(TestCompiledModelDriverNoStart, self).teardown()


class TestCompiledModelDriver(TestCompiledModelParam,
                              parent.TestModelDriver):
    r"""Test runner for CompiledModelDriver."""
    pass



import os
from yggdrasil.examples.tests import ExampleTstBase


class TestExampleHello(ExampleTstBase):
    r"""Test the Hello example."""

    example_name = 'hello'

    @property
    def input_files(self):
        r"""Input file."""
        return [os.path.join(self.yamldir, 'Input', 'input.txt')]
    
    @property
    def output_files(self):
        r"""Output file."""
        return [os.path.join(self.tempdir, 'output_hello.txt')]



import os
import sys
from yggdrasil import tools, platform
from yggdrasil.tests import YggTestClass, assert_equal


def test_get_conda_prefix():
    r"""Test get_conda_prefix."""
    tools.get_conda_prefix()

    
def test_get_conda_env():
    r"""Test get_conda_env."""
    tools.get_conda_env()
    

def test_get_supported():
    r"""Test get_supported_<platforms/lang/comm>."""
    tools.get_supported_platforms()
    tools.get_supported_lang()
    tools.get_supported_comm()


def test_get_installed():
    r"""Test get_installed_<lang/comm>."""
    tools.get_installed_lang()
    tools.get_installed_comm()


def test_which():
    r"""Test location of executable."""
    assert(tools.which(sys.executable) is not None)
    if platform._is_win:  # pragma: windows
        assert(tools.which('python.exe') is not None)
    else:
        assert(tools.which('python') is not None)
    assert(tools.which('invalid') is None)


def test_locate_path():
    r"""Test file location."""
    # Search for current file
    fdir, fname = os.path.split(__file__)
    basedir = os.path.dirname(fdir)
    fpath = tools.locate_path(fname, basedir=basedir)
    assert(fpath)
    assert(__file__ in fpath)
    # assert_equal(__file__, fpath)
    # Search for invalid file
    fname = 'invalid_file.ext'
    fpath = tools.locate_path(fname, basedir=basedir)
    assert(not fpath)


def test_popen_nobuffer():
    r"""Test open of process without buffer."""
    ans = os.getcwd()  # + '\n'
    # ans = backwards.as_bytes(ans)
    # Test w/o shell
    if platform._is_win:  # pragma: windows
        args = ['cmd', '/c', 'cd']
    else:
        args = ['pwd']
    p = tools.popen_nobuffer(args)
    out, err = p.communicate()
    res = out.decode('utf-8').splitlines()[0]
    assert_equal(res, ans)
    # Test w/ shell
    if platform._is_win:  # pragma: windows
        args = 'cd'
    else:
        args = 'pwd'
    p = tools.popen_nobuffer(args, shell=True)
    out, err = p.communicate()
    res = out.decode('utf-8').splitlines()[0]
    assert_equal(res, ans)


def test_eval_kwarg():
    r"""Ensure strings & objects properly evaluated."""
    vals = [None, True, False, ['one', 'two'], 'one']
    for v in vals:
        assert_equal(tools.eval_kwarg(v), v)
        assert_equal(tools.eval_kwarg(str(v)), v)
    assert_equal(tools.eval_kwarg("'one'"), 'one')
    assert_equal(tools.eval_kwarg('"one"'), 'one')


class TestYggClass(YggTestClass):
    r"""Test basic behavior of YggTestClass."""

    _cls = 'YggClass'
    _mod = 'yggdrasil.tools'

    def __init__(self, *args, **kwargs):
        super(TestYggClass, self).__init__(*args, **kwargs)
        self.namespace = 'TESTING_%s' % self.uuid
        self.attr_list += ['name', 'sleeptime', 'longsleep', 'timeout']
        self._inst_kwargs = {'timeout': self.timeout,
                             'sleeptime': self.sleeptime}
        self.debug_flag = False

    def test_attributes(self):
        r"""Assert that the driver has all of the required attributes."""
        for a in self.attr_list:
            if not hasattr(self.instance, a):  # pragma: debug
                raise AttributeError("Driver does not have attribute %s" % a)

    def test_prints(self):
        r"""Test logging at various levels."""
        self.instance.display(1)
        self.instance.info(1)
        self.instance.debug(1)
        self.instance.verbose_debug(1)
        self.instance.critical(1)
        self.instance.warning(1)
        self.instance.error(1)
        self.instance.exception(1)
        try:
            raise Exception("Test exception")
        except Exception:
            self.instance.exception(1)
        self.instance.printStatus()
        self.instance.special_debug(1)
        self.instance.suppress_special_debug = True
        self.instance.special_debug(1)
        self.instance.suppress_special_debug = False
        self.instance.interface_info(1)
        self.instance.language_info([])(1)
        self.instance.language_info('python')(1)

    def test_timeout(self):
        r"""Test functionality of timeout."""
        # Test w/o timeout
        self.instance.start_timeout(10, key='fake_key')
        assert(not self.instance.check_timeout(key='fake_key'))
        # Test errors
        self.assert_raises(KeyError, self.instance.start_timeout,
                           0.1, key='fake_key')
        self.instance.stop_timeout(key='fake_key')
        self.assert_raises(KeyError, self.instance.check_timeout)
        self.assert_raises(KeyError, self.instance.check_timeout, key='fake_key')
        self.assert_raises(KeyError, self.instance.stop_timeout, key='fake_key')
        # Test w/ timeout
        T = self.instance.start_timeout(0.001)  # self.instance.sleeptime)
        while not T.is_out:
            self.instance.sleep()
        assert(self.instance.check_timeout())
        self.instance.stop_timeout(quiet=True)



import unittest
import zmq
import copy
from yggdrasil import platform
from yggdrasil.tests import assert_raises, assert_equal
from yggdrasil.communication import new_comm
from yggdrasil.communication.tests import test_AsyncComm
from yggdrasil.communication import ZMQComm, IPCComm


_zmq_installed = ZMQComm.ZMQComm.is_installed(language='python')
_ipc_installed = IPCComm.IPCComm.is_installed(language='python')


@unittest.skipIf(not _zmq_installed, "ZMQ library not installed")
def test_get_socket_type_mate():
    r"""Test socket type matching."""
    for s, r in ZMQComm._socket_type_pairs:
        assert_equal(ZMQComm.get_socket_type_mate(s), r)
        assert_equal(ZMQComm.get_socket_type_mate(r), s)
    assert_raises(ValueError, ZMQComm.get_socket_type_mate, 'INVALID')


@unittest.skipIf(not _zmq_installed, "ZMQ library not installed")
def test_format_address():
    r"""Test format/parse of address."""
    protocol = 'tcp'
    host = '127.0.0.1'
    port = 5555
    address = ZMQComm.format_address(protocol, host, port)
    result = ZMQComm.parse_address(address)
    assert_equal(result['protocol'], protocol)
    assert_equal(result['host'], host)
    assert_equal(result['port'], port)
    assert_raises(ValueError, ZMQComm.parse_address, 'INVALID')
    assert_raises(ValueError, ZMQComm.parse_address, 'INVALID://')


@unittest.skipIf(not _zmq_installed, "ZMQ library not installed")
def test_invalid_protocol():
    r"""Test raise of an error in the event of an invalid protocol."""
    assert_raises(ValueError, new_comm, 'test_invalid_protocol',
                  comm='ZMQComm', protocol='invalid')


@unittest.skipIf(not _zmq_installed, "ZMQ library not installed")
@unittest.skipIf(platform._is_mac, "Testing on MacOS")
@unittest.skipIf(platform._is_win, "Testing on Windows")
def test_error_on_send_open_twice():
    r"""Test creation of the same send socket twice for an error."""
    for s, r in ZMQComm._socket_type_pairs:
        # Send comm
        name1 = 'test_%s' % s
        comm1 = new_comm(name1 + '_1', comm='ZMQComm', socket_type=s,
                         dont_open=True, socket_action='bind')
        assert_raises(zmq.ZMQError, ZMQComm.ZMQComm,
                      name1 + '_2', socket_type=s,
                      address=comm1.opp_address, socket_action='bind')
        comm1.close()

        
@unittest.skipIf(not _zmq_installed, "ZMQ library not installed")
class TestZMQComm(test_AsyncComm.TestAsyncComm):
    r"""Test for ZMQComm communication class."""

    comm = 'ZMQComm'
    attr_list = (copy.deepcopy(test_AsyncComm.TestAsyncComm.attr_list)
                 + ['context', 'socket', 'socket_type_name',
                    'socket_type', 'protocol', 'host', 'port'])
    protocol = None
    socket_type = None

    @property
    def description_prefix(self):
        r"""String prefix to prepend docstr test message with."""
        return '%s(%s, %s)' % (self.comm, self.protocol, self.socket_type)

    @property
    def send_inst_kwargs(self):
        r"""Keyword arguments for send instance."""
        out = super(TestZMQComm, self).send_inst_kwargs
        out['protocol'] = self.protocol
        out['socket_type'] = self.socket_type
        return out

    def test_send_recv_nolimit(self):
        r"""Send/recv of large message."""
        if self.__class__ != TestZMQComm:
            raise unittest.SkipTest('Only test once')
        super(TestZMQComm, self).test_send_recv_nolimit()

    def test_eof_no_close(self):
        r"""Test send/recv of EOF message with no close."""
        if self.__class__ != TestZMQComm:
            raise unittest.SkipTest('Only test once')
        super(TestZMQComm, self).test_eof_no_close()
        
    
# Tests for server/client
class TestZMQComm_client(TestZMQComm):
    r"""Test for ZMQComm communication class for client/server."""

    @property
    def send_inst_kwargs(self):
        r"""Keyword arguments for send instance."""
        out = super(TestZMQComm_client, self).send_inst_kwargs
        out['is_client'] = True
        return out

    
# Tests for all the supported protocols
class TestZMQCommINPROC(TestZMQComm):
    r"""Test for ZMQComm communication class with INPROC socket."""

    protocol = 'inproc'

    
class TestZMQCommTCP(TestZMQComm):
    r"""Test for ZMQComm communication class with TCP socket."""

    protocol = 'tcp'

    
@unittest.skipIf(not _ipc_installed, "IPC library not installed")
class TestZMQCommIPC(TestZMQComm):
    r"""Test for ZMQComm communication class with IPC socket."""

    protocol = 'ipc'


class TestZMQCommIPC_client(TestZMQComm_client, TestZMQCommIPC):
    r"""Test for ZMQComm communication class with IPC socket."""
    pass
    

# Unsupported
# class TestZMQCommUDP(TestZMQComm):
#     r"""Test for ZMQComm communication class with UDP socket."""

#     protocol = 'udp'


# class TestZMQCommPGM(TestZMQComm):
#     r"""Test for ZMQComm communication class with PGM socket."""

#     protocol = 'pgm'

    
# class TestZMQCommEPGM(TestZMQComm):
#     r"""Test for ZMQComm communication class with EPGM socket."""

#     protocol = 'epgm'


# Tests for all the socket types
class TestZMQCommPAIR(TestZMQComm):
    r"""Test for ZMQComm communication class with PAIR/PAIR socket."""

    socket_type = 'PAIR'

    
class TestZMQCommPUSH(TestZMQComm):
    r"""Test for ZMQComm communication class with PUSH/PULL socket."""

    socket_type = 'PUSH'

    
class TestZMQCommPUSH_INPROC(TestZMQCommINPROC):
    r"""Test for ZMQComm communication class with INPROC PUSH/PULL socket."""

    socket_type = 'PUSH'

    
class TestZMQCommPUB(TestZMQComm):
    r"""Test for ZMQComm communication class with PUB/SUB socket."""

    socket_type = 'PUB'


class TestZMQCommREQ(TestZMQComm):
    r"""Test for ZMQComm communication class with REP/REQ socket."""

    socket_type = 'REQ'


class TestZMQCommROUTER(TestZMQComm):
    r"""Test for ZMQComm communication class with DEALER/ROUTER socket."""

    socket_type = 'ROUTER'

    def setup(self, *args, **kwargs):
        r"""Initialize comm object pair with sleep after setup to ensure
        dealer has connected."""
        kwargs['sleep_after_connect'] = True
        super(TestZMQCommROUTER, self).setup(*args, **kwargs)

    def test_router_recv(self):
        r"""Test router receipt of message from the dealer with an identity."""
        self.do_send_recv(reverse_comms=True, send_kwargs=dict(
            identity=self.recv_instance.dealer_identity))


# @unittest.skipIf(_zmq_installed, "ZMQ library installed")
# def test_not_running():
#     r"""Test raise of an error if a ZMQ library is not installed."""
#     comm_kwargs = dict(comm='ZMQComm', direction='send', reverse_names=True)
#     assert_raises(RuntimeError, new_comm, 'test', **comm_kwargs)



from yggdrasil.communication.FileComm import FileComm


class AsciiFileComm(FileComm):
    r"""Class for handling I/O from/to a file on disk.

    Args:
        name (str): The environment variable where communication address is
            stored.
        comment (str, optional): String indicating a comment. If 'read_meth'
            is 'readline' and this is provided, lines starting with a comment
            will be skipped.
        **kwargs: Additional keywords arguments are passed to parent class.

    """

    _filetype = 'ascii'
    _schema_subtype_description = ('This file is read/written as encoded text '
                                   'one line at a time.')

    def _init_before_open(self, **kwargs):
        r"""Get absolute path and set attributes."""
        super(AsciiFileComm, self)._init_before_open(**kwargs)
        self.read_meth = 'readline'

    @classmethod
    def get_testing_options(cls, **kwargs):
        r"""Method to return a dictionary of testing options for this class.

        Returns:
            dict: Dictionary of variables to use for testing. Key/value pairs:
                kwargs (dict): Keyword arguments for comms tested with the
                    provided content.
                send (list): List of objects to send to test file.
                recv (list): List of objects that will be received from a test
                    file that was sent the messages in 'send'.
                contents (bytes): Bytes contents of test file created by sending
                    the messages in 'send'.

        """
        kwargs['read_meth'] = 'readline'
        return super(AsciiFileComm, cls).get_testing_options(**kwargs)



from yggdrasil.metaschema.datatypes import (
    transform_type, encode_data, encode_data_readable, decode_data)
from yggdrasil.metaschema.datatypes.MetaschemaType import MetaschemaType


class AnyMetaschemaType(MetaschemaType):
    r"""Type associated with a scalar."""

    name = 'any'
    description = 'A type allowing any value that is expresible in some type.'
    properties = ['temptype']
    metadata_properties = ['temptype']
    python_types = (object, )
    cross_language_support = False

    @classmethod
    def validate(cls, *args, **kwargs):
        r"""Validate an object to check if it could be of this type. For this
        type, the returned boolean will always be True."""
        return True
        
    @classmethod
    def issubtype(cls, t):
        r"""Determine if this type is a subclass of the provided type.

        Args:
            t (str): Type name to check against.

        Returns:
            bool: True if this type is a subtype of the specified type t.

        """
        return True
    
    @classmethod
    def get_temptype(cls, typedef):
        r"""Extract temporary type from type definition.
        
        Args:
            typedef (dict): Type definition containing temporary type under the
                temptype key.

        Returns:
            dict: Temporary type definition or None if not present.

        """
        out = None
        if isinstance(typedef, dict):
            out = typedef.get('temptype', None)
        return out
        
    @classmethod
    def encode_data(cls, obj, typedef):
        r"""Encode an object's data.

        Args:
            obj (object): Object to encode.
            typedef (dict): Type definition that should be used to encode the
                object.

        Returns:
            string: Encoded object.

        """
        return encode_data(obj, cls.get_temptype(typedef))

    @classmethod
    def encode_data_readable(cls, obj, typedef):
        r"""Encode an object's data in a readable format that may not be
        decoded in exactly the same way.

        Args:
            obj (object): Object to encode.
            typedef (dict): Type definition that should be used to encode the
                object.

        Returns:
            string: Encoded object.

        """
        return encode_data_readable(obj, cls.get_temptype(typedef))

    @classmethod
    def decode_data(cls, obj, typedef):
        r"""Decode an object.

        Args:
            obj (string): Encoded object to decode.
            typedef (dict): Type definition that should be used to decode the
                object.

        Returns:
            object: Decoded object.

        """
        return decode_data(obj, cls.get_temptype(typedef))

    @classmethod
    def transform_type(cls, obj, typedef=None):
        r"""Transform an object based on type info.

        Args:
            obj (object): Object to transform.
            typedef (dict, optional): Type definition that should be used to
                transform the object. Defaults to None.

        Returns:
            object: Transformed object.

        """
        return transform_type(obj, cls.get_temptype(typedef))



import copy
from yggdrasil.languages.Python import test_YggInterface


title_replace = ['TestYggInput', 'TestYggInputMatlab',
                 'TestYggRpcClient', 'TestYggRpcClientMatlab',
                 'TestYggAsciiFileInput', 'TestYggAsciiTableInput',
                 'TestYggAsciiTableOutputMatlab',
                 'TestYggAsciiArrayInput', 'TestYggPickleInput',
                 'TestYggPandasInput', 'TestYggPlyInput',
                 'TestYggObjInput']
for x in copy.deepcopy(title_replace):
    if 'Input' in x:
        title_replace.append(x.replace('Input', 'Output'))
    elif 'Client' in x:
        title_replace.append(x.replace('Client', 'Server'))
for x in title_replace:
    globals()[x.replace('Ygg', 'Cis')] = type(
        x.replace('Ygg', 'Cis'), (getattr(test_YggInterface, x), ),
        {'_mod': 'yggdrasil.interface.CisInterface',
         '_cls': getattr(test_YggInterface, x)._cls.replace('Ygg', 'Cis')})



import re
import copy
import numpy as np
import pandas
from yggdrasil import backwards, platform, units
try:
    if not backwards.PY2:  # pragma: Python 3
        from astropy.io import ascii as apy_ascii
        from astropy.table import Table as apy_Table
        _use_astropy = True
    else:  # pragma: Python 2
        apy_ascii, apy_Table = None, None
        _use_astropy = False
except ImportError:  # pragma: no cover
    apy_ascii, apy_Table = None, None
    # print("astropy is not installed, reading/writing as an array will be "
    #       + "disabled. astropy can be installed using 'pip install astropy'.")
    _use_astropy = False


_fmt_char = b'%'
_default_comment = b'# '
_default_delimiter = b'\t'
_default_newline = b'\n'


def extract_formats(fmt_str):
    r"""Locate format codes within a format string.

    Args:
        fmt_str (str, bytes): Format string.

    Returns:
        list: List of identified format codes.

    """
    fmt_regex = (
        "%(?:\\d+\\$)?[+-]?(?:[ 0]|\'.{1})?-?\\d*(?:\\.\\d+)?"
        + "[lhjztL]*(?:64)?[bcdeEufFgGosxXi]"
        + "(?:%(?:\\d+\\$)?[+-](?:[ 0]|\'.{1})?-?\\d*(?:\\.\\d+)?"
        + "[lhjztL]*[eEfFgG]j)?")
    out = re.findall(fmt_regex, backwards.as_str(fmt_str))
    out = [backwards.match_stype(fmt_str, f) for f in out]
    return out
    

def nptype2cformat(nptype, asbytes=False):
    r"""Convert a numpy data type to a c format string.

    Args:
        nptype (str or numpy.dtype): Numpy data type that should be converted.
        asbytes (bool, optional): If True, the returned format codes will be
            in bytes format (str for Python 2, bytes for Python 3). Defaults to
            False.

    Returns:
        str, list: Corresponding c format specification string or list of
            format strings for complex types..

    Raises:
        TypeError: If nptype is not a string or numpy.dtype.
        ValueError: If a matching format string cannot be determined.

    """
    if isinstance(nptype, np.dtype):
        t = nptype
    elif isinstance(nptype, str):
        t = np.dtype(nptype)
    else:
        raise TypeError("Input must be a string or a numpy.dtype")
    if len(t) > 0:
        out = [nptype2cformat(t[n], asbytes=asbytes) for n in t.names]
        return out
    if t in [np.dtype(x) for x in ["float_", "float16", "float32", "float64"]]:
        cfmt = "%g"  # Ensures readability
    elif t in [np.dtype(x) for x in ["complex_", "complex64", "complex128"]]:
        cfmt = "%g%+gj"
    elif t == np.dtype("int8"):
        cfmt = "%hhd"
    elif t == np.dtype("short"):
        cfmt = "%hd"
    elif t == np.dtype("intc"):
        cfmt = "%d"
    elif t == np.dtype("int_"):
        cfmt = "%ld"
    # elif t == np.dtype("int64"):
    #     if platform._is_win:
    #         cfmt = "%l64d"
    #     else:  # pragma: no cover
    #         cfmt = "%ld"
    elif t == np.dtype("longlong"):
        # On windows C long is 32bit and long long is 64bit
        if platform._is_win:  # pragma: windows
            cfmt = "%l64d"
        else:  # pragma: no cover
            cfmt = "%lld"
    elif t == np.dtype("uint8"):
        cfmt = "%hhu"
    elif t == np.dtype("ushort"):
        cfmt = "%hu"
    elif t == np.dtype("uintc"):
        cfmt = "%u"
    elif t == np.dtype("uint64"):  # Platform dependent
        if platform._is_win:  # pragma: windows
            cfmt = "%l64u"
        else:
            cfmt = "%lu"
    elif t == np.dtype("ulonglong"):  # pragma: no cover
        # On windows C unsigned long is 32bit and unsigned long long is 64bit
        if platform._is_win:  # pragma: windows
            cfmt = "%l64u"
        else:
            cfmt = "%llu"
    elif np.issubdtype(t, np.dtype("S")):
        if t.itemsize == 0:
            cfmt = '%s'
        else:
            cfmt = "%" + str(t.itemsize) + "s"
    elif np.issubdtype(t, np.dtype("U")):
        if t.itemsize == 0:
            cfmt = '%s'
        else:
            cfmt = "%" + str(t.itemsize) + "s"
    else:
        raise ValueError("No format specification string for dtype %s" % t)
    # Short and long specifiers not supported by python scanf
    # cfmt = cfmt.replace("h", "")
    # cfmt = cfmt.replace("l", "")
    if asbytes:
        cfmt = backwards.as_bytes(cfmt)
    return cfmt


def cformat2nptype(cfmt, names=None):
    r"""Convert a c format string to a numpy data type.

    Args:
        cfmt (str, bytes): c format that should be translated.
        names (list, optional): Names that should be assigned to fields in the
            format string if there is more than one. If not provided, names
            are generated based on the order of the format codes.

    Returns:
        np.dtype: Corresponding numpy data type.

    Raises:
        TypeError: if cfmt is not a string.
        ValueError: If the c format does not begin with '%'.
        ValueError: If the c format does not contain type info.
        ValueError: If the c format cannot be translated to a numpy datatype.

    """
    # TODO: this may fail on 32bit systems where C long types are 32 bit
    if not (isinstance(cfmt, list) or isinstance(cfmt, backwards.string_types)):
        raise TypeError("Input must be a string, bytes string, or list, not %s" %
                        type(cfmt))
    if isinstance(cfmt, backwards.string_types):
        fmt_list = extract_formats(backwards.as_str(cfmt))
        if len(fmt_list) == 0:
            raise ValueError("Could not locate any format codes in the "
                             + "provided format string (%s)." % cfmt)
    else:
        fmt_list = cfmt
    nfmt = len(fmt_list)
    if nfmt == 1:
        cfmt_str = fmt_list[0]
    else:
        dtype_list = [cformat2nptype(f) for f in fmt_list]
        if names is None:
            names = ['f%d' % i for i in range(nfmt)]
        elif len(names) != nfmt:
            raise ValueError("Number of names does not match the number of fields.")
        else:
            names = [backwards.as_str(n) for n in names]
        out = np.dtype(dict(names=names, formats=dtype_list))
        # out = np.dtype([(n, d) for n, d in zip(names, dtype_list)])
        return out
    out = None
    if cfmt_str[-1] in ['j']:
        out = 'complex128'
    elif cfmt_str[-1] in ['f', 'F', 'e', 'E', 'g', 'G']:
        # if 'hh' in cfmt_str:
        #     out = 'float8'
        # elif cfmt_str[-2] == 'h':
        #     out = 'float16'
        # elif 'll' in cfmt_str:
        #     out = 'longfloat'
        # elif cfmt_str[-2] == 'l':
        #     out = 'double'
        # else:
        #     out = 'single'
        out = 'float64'
    elif cfmt_str[-1] in ['d', 'i']:
        if 'hh' in cfmt_str:  # short short, single char
            out = 'int8'
        elif cfmt_str[-2] == 'h':  # short
            out = 'short'
        elif ('ll' in cfmt_str) or ('l64' in cfmt_str):
            out = 'longlong'  # long long
        elif cfmt_str[-2] == 'l':
            out = 'int_'  # long (broken in python)
        else:
            out = 'intc'  # int, platform dependent
    elif cfmt_str[-1] in ['u', 'o', 'x', 'X']:
        if 'hh' in cfmt_str:  # short short, single char
            out = 'uint8'
        elif cfmt_str[-2] == 'h':  # short
            out = 'ushort'
        elif ('ll' in cfmt_str) or ('l64' in cfmt_str):
            out = 'ulonglong'  # long long
        elif cfmt_str[-2] == 'l':
            out = 'uint64'  # long (broken in python)
        else:
            out = 'uintc'  # int, platform dependent
    elif cfmt_str[-1] in ['c', 's']:
        lstr = cfmt_str[1:-1]
        if lstr:
            lint = int(lstr)
        else:
            lint = 0
        lsiz = lint * np.dtype(backwards.np_dtype_str + '1').itemsize
        out = '%s%d' % (backwards.np_dtype_str, lsiz)
    else:
        raise ValueError("Could not find match for format str %s" % cfmt)
    return np.dtype(out)


def cformat2pyscanf(cfmt):
    r"""Convert a c format specification string to a version that the
    python scanf module can use.

    Args:
        cfmt (str, bytes, list): C format specification string or list of format
            strings.

    Returns:
        str, bytes, list: Version of cfmt or list of cfmts that can be parsed by
            scanf.

    Raises:
        TypeError: If cfmt is not a bytes/str/list.
        ValueError: If there are not an format codes in the format string.

    """
    if not (isinstance(cfmt, list) or isinstance(cfmt, backwards.string_types)):
        raise TypeError("Input must be a string, bytes string, or list, not %s" %
                        type(cfmt))
    if isinstance(cfmt, list):
        return [cformat2pyscanf(f) for f in cfmt]
    cfmt_out = backwards.as_str(cfmt)
    fmt_list = extract_formats(cfmt_out)
    if len(fmt_list) == 0:
        raise ValueError("Could not locate any format codes in the "
                         + "provided format string (%s)." % cfmt)
    for cfmt_str in fmt_list:
        # Hacky, but necessary to handle concatenation of a single byte
        if cfmt_str[-1] == 's':
            out = '%s'
        else:
            out = cfmt_str
        # if cfmt_str[-1] == 'j':
        #     # Handle complex format specifier
        #     out = '%g%+gj'
        # else:
        #     out = backwards.as_str(_fmt_char)
        #     out += cfmt_str[-1]
        #     out = out.replace('h', '')
        #     out = out.replace('l', '')
        #     out = out.replace('64', '')
        cfmt_out = cfmt_out.replace(cfmt_str, out, 1)
    cfmt_out = backwards.match_stype(cfmt, cfmt_out)
    return cfmt_out


def format_message(args, fmt_str):
    r"""Format a message from a list of arguments and a format string.

    Args:
        args (list, obj): List of arguments or single argument that should be
            formatted using the format string.
        fmt_str (str, bytes): Format string that should be used to format the
            arguments.

    Returns:
        str, bytes: Formatted message. The type will match the type of the
            fmt_str.

    Raises:
        RuntimeError: If the number of arguments does not match the number of
            format fields.

    """
    if not isinstance(args, (tuple, list)):
        args = (args, )
    nfmt = len(extract_formats(fmt_str))
    args_ = []
    if len(args) < nfmt:
        raise RuntimeError("Number of arguments (%d) does not match " % len(args)
                           + "number of format fields (%d)." % nfmt)
    for a0 in args:
        a = units.get_data(a0)
        if np.iscomplexobj(a):
            args_ += [a.real, a.imag]
        else:
            args_.append(a)
    out = backwards.format_bytes(fmt_str, tuple(args_))
    return out


def process_message(msg, fmt_str):
    r"""Extract python objects from a message using a format string.

    Args:
        msg (str, bytes): Message that should be parsed.
        fmt_str (str, bytes): Format string that should be used to parse the
            message using scanf.

    Returns:
        tuple: Variables extracted from the message.

    Raises:
        TypeError: If the message is not a string or bytes string type.
        ValueError: If the expected number of variables cannot be extracted
            from the message.

    """
    if not isinstance(msg, backwards.string_types):
        raise TypeError("Message must be a string or bytes string type.")
    nfmt = len(extract_formats(fmt_str))
    py_fmt_str = cformat2pyscanf(fmt_str)
    args = backwards.scanf_bytes(py_fmt_str, msg)
    if args is None:
        nargs = 0
    else:
        nargs = len(args)
        if len(args) > 1:
            dtype = cformat2nptype(fmt_str)
            dtype_list = [dtype[i] for i in range(nargs)]
            args = tuple([np.array([a], idtype)[0] for
                          a, idtype in zip(args, dtype_list)])
    if nargs != nfmt:
        raise ValueError("%d arguments were extracted, " % nargs
                         + "but format string expected %d." % nfmt)
    return args


def combine_flds(arrs, dtype=None):
    r"""Combine a list of arrays as fields in a structured array.

    Args:
        arrs (list): List of arrays, one for each field in a structured array.
        dtype (np.dtype, optional): Data type that resulting array should have.
            If not provided, it is determined from the types of the input arrays.

    Returns:
        np.ndarray: Structured numpy array with fields assigned from the
            provided arrays.

    Raises:
        ValueError: If the number of arrays does not match the number of fields
            in the provided data type.
        ValueError: If any of the provided arrays does not have the same number
            of elements as the first array.

    """
    # Get dtype
    nflds = len(arrs)
    shape = arrs[0].shape
    if dtype is None:
        names = ['f%d' % i for i in range(nflds)]
        dtype = np.dtype([(n, a.dtype) for n, a in zip(names, arrs)])
    # Check number of fields and array shapes
    if len(dtype) != nflds:
        raise ValueError("dtype has %d fields, but %d arrays were provided." %
                         (len(dtype), nflds))
    for i, a in enumerate(arrs):
        if a.shape != shape:
            raise ValueError("Shape of array %d (%s) does " % (i, a.shape)
                             + "match shape of 1st array (%s)." % shape)
    # Combine arrays
    out = np.empty(shape, dtype=dtype)
    for i, a in enumerate(arrs):
        out[dtype.names[i]] = a
    return out


def combine_eles(arrs, dtype=None):
    r"""Combine arrays as elements in a structure numpy array.

    Args:
        arrs (list): List of numpy arrays or lists specifying the values of
            fields in elements of a structured numpy array.
        dtype (np.dtype, optional): Data type that the resulting array should
            have. If not provided, it is determined from the provided arrays.

    Returns:
        np.ndarray: Structured numpy array with information from the provided
            arrays/lists as elements.

    Raises:
        ValueError: If the number of fields in any of the provided arrays/lists
            does not match the number of fields in the provided/created dtype.

    """
    neles = len(arrs)
    # Check types of arrays
    any_arrays = False
    all_arrays = True
    for i, a in enumerate(arrs):
        if isinstance(a, (np.ndarray, np.void, list, tuple)):
            any_arrays = True
        else:
            all_arrays = False
    if not all_arrays:
        if not any_arrays:
            arrs = [arrs]
            neles = 1
        else:
            raise TypeError("Elements must be arrays, lists or tuples. "
                            + "Element %d has type %s." % (i, type(a)))
    # Check that there are enough elements for each data type
    if dtype is None:
        if isinstance(arrs[0], (np.ndarray, np.void)):
            nflds = len(arrs[0].dtype)
        else:
            nflds = len(arrs[0])
    else:
        nflds = len(dtype)
    for i, a in enumerate(arrs):
        if isinstance(a, (np.ndarray, np.void)):
            if len(a.dtype) != nflds:
                raise ValueError("Element %d has dtype %s, but " % (i, a.dtype)
                                 + "%d fields are expected." % nflds)
        else:
            if len(a) != nflds:
                raise ValueError("Element %d has %d values, but " % (i, len(a))
                                 + "%d fields are expected." % nflds)
    # Get data type
    if dtype is None:
        if isinstance(arrs[0], (np.ndarray, np.void)):
            dtype = arrs[0].dtype
        elif isinstance(arrs[0], (list, tuple)):
            names = None
            for iele in arrs:
                if isinstance(iele, (np.ndarray, np.void)):
                    names = iele.dtype.names
                    break
            nflds = len(arrs[0])
            dtype_list = []
            for i, a in enumerate(arrs[0]):
                dtype_str = np.dtype(type(a)).str
                if 'S' in dtype_str:
                    max_len = 0
                    for iele in arrs:
                        if isinstance(iele, (np.ndarray, np.void)):
                            n = iele.dtype.names[i]
                            max_len = max(max_len, len(iele[n]))
                        else:
                            max_len = max(max_len, len(iele[i]))
                    dtype_str = 'S%d' % max_len
                dtype_list.append(np.dtype(dtype_str))
            if names is None:
                names = ['f%d' % i for i in range(nflds)]
            dtype = np.dtype({'names': names, 'formats': dtype_list})
    # Combine rows
    out = np.empty(neles, dtype=dtype)
    for i, a in enumerate(arrs):
        if isinstance(a, (np.ndarray, np.void)):
            out[i] = a
        else:
            for j, n in enumerate(dtype.names):
                out[n][i] = a[j]
    return out


def consolidate_array(arrs, dtype=None):
    r"""Consolidate arrays either as fields or elements in a structured array.

    Args:
        arrs (np.ndarray, list, tuple): Array or list of arrays that should be
            reformatted/consolidated as a structured array.
        dtype (np.dtype, optional): Data type that the resulting array should
            have. If not provided, it is determined from the input.

    Returns:
        np.ndarray: Structured array of the specified type with values taken
            from the provided array(s).

    Raises:
        ValueError: If an unstructured array is provided, but the last dimension
            of the array does not have the proper size to fill the fields in
            the specified data type.
        ValueError: If a structured array is provided, but it does not have a
            data type compatible with the specified data type.
        TypeError: If 'arrs' is not a numpy array, list, or tuple.

    """
    if isinstance(arrs, np.ndarray):
        if (dtype is None) or (dtype == arrs.dtype):
            out = arrs
        else:
            if len(arrs.dtype) == 0:
                if arrs.shape[-1] != len(dtype):
                    raise ValueError("The last dimension of the input array "
                                     + "(%d) " % arrs.shape[-1]
                                     + "dosn't match the number of fields in "
                                     + "the dtype (%d)." % len(dtype))
                out = np.empty(arrs.shape[:-1], dtype=dtype)
                for i in range(arrs.shape[-1]):
                    out[dtype.names[i]] = arrs[..., i]
            elif len(arrs.dtype) == len(dtype):
                out = np.empty(arrs.shape, dtype=dtype)
                for n1, n2 in zip(arrs.dtype.names, dtype.names):
                    out[n2] = arrs[n1]
            else:
                raise ValueError("The input array data type (%s) " % arrs.dtype
                                 + "is not compatible with the specified "
                                 + "data type (%s)." % dtype)
    elif isinstance(arrs, (list, tuple)):
        if isinstance(arrs[0], (np.ndarray, np.void)):
            if len(arrs[0].dtype) > 1:
                out = combine_eles(arrs, dtype=dtype)
            else:
                out = combine_flds(arrs, dtype=dtype)
        else:
            out = combine_eles(arrs, dtype=dtype)
    else:
        raise TypeError("Input 'arrs' must be a numpy array, list, or tuple.")
    return out


def format2table(fmt_str):
    r"""Get table information from the table format string.

    Args:
        fmt_str (str, bytes): Format string that describes how the table
            columns are formatted.

    Returns:
        dict: Table information parameters.

    Raises:
        RuntimeError: If the column separator is not the same between all
            of the format codes.

    """
    out = {}
    out['fmts'] = extract_formats(fmt_str)
    if len(out['fmts']) == 0:
        return out
    seps = []
    comment, fmt_rem = fmt_str.split(out['fmts'][0], 1)
    if comment:
        out['comment'] = comment
    for f in out['fmts'][1:]:
        isep, fmt_rem = fmt_rem.split(f, 1)
        seps.append(isep)
    if fmt_rem:
        out['newline'] = fmt_rem
    seps = set(seps)
    if len(seps) == 0:
        out['delimiter'] = _default_delimiter
    elif len(seps) == 1:
        out['delimiter'] = list(seps)[0]
    elif len(seps) > 1:
        for s in seps:
            if not s.isspace():
                raise RuntimeError("There is more than one column separator (%s)." % seps)
        out['delimiter'] = min(seps, key=len)
    return out


def table2format(fmts=[], delimiter=None, newline=None, comment=None):
    r"""Create a format string from table information.

    Args:
        fmts (list, optional): List of format codes for each column. Defaults to
            [].
        delimiter (bytes, optional): String used to separate columns. Defaults
            to _default_delimiter.
        newline (bytes, optional): String used to indicate the end of a table
            line. Defaults to _default_newline.
        comment (bytes, optional): String that should be prepended to the format
            string to indicate a comment. Defaults to _default_comment.

    Returns:
        str, bytes: Table format string.

    """
    if delimiter is None:
        delimiter = _default_delimiter
    if newline is None:
        newline = _default_newline
    if comment is None:
        comment = _default_comment
    if isinstance(fmts, np.dtype):
        fmts = nptype2cformat(fmts)
    bytes_fmts = [backwards.as_bytes(f) for f in fmts]
    fmt_str = comment + delimiter.join(bytes_fmts) + newline
    return fmt_str


# def unicode_dtype(old_dtype):
#     r"""Convert a dtype to use unicode.

#     Args:
#         old_dtype (np.dtype): Numpy data type.

#     Returns:
#         np.dtype: Numpy dtype with bytes replaced with unicode.

#     """
#     if backwards.PY2:  # pragma: Python 2
#         new_dtype = old_dtype
#     else:  # pragma: Python 3
#         ntype = len(old_dtype)
#         if ntype > 0:
#             names = old_dtype.names
#             types = [unicode_dtype(old_dtype[i]) for i in range(ntype)]
#             new_type = dict(formats=types, names=names)
#             new_dtype = np.dtype(new_type)
#         else:
#             if np.issubdtype(old_dtype, np.dtype('S')):
#                 front, width = old_dtype.str.split('S')
#                 new_dtype = np.dtype(front + 'U' + width)
#             else:
#                 new_dtype = old_dtype
#     return new_dtype


def array_to_table(arrs, fmt_str, use_astropy=False):
    r"""Serialize an array as an ASCII table.

    Args:
        arrs (np.ndarray, list, tuple): Structured array or list/tuple of
            arrays that contain table information.
        fmt_str (str, bytes): Format string that should be used to structure
            the ASCII array.
        use_astropy (bool, optional): If True, astropy will be used to format
            the table if it is installed. Defaults to False.

    Returns:
        bytes: ASCII table.

    """
    if not _use_astropy:
        use_astropy = False
    dtype = cformat2nptype(fmt_str)
    info = format2table(fmt_str)
    comment = info.get('comment', None)
    if comment is not None:
        fmt_str = fmt_str.split(comment, 1)[-1]
    arr1 = consolidate_array(arrs, dtype=dtype)
    if use_astropy:
        fd = backwards.StringIO()
        table = apy_Table(arr1)
        delimiter = info['delimiter']
        delimiter = backwards.as_str(delimiter)
        apy_ascii.write(table, fd, delimiter=delimiter,
                        format='no_header')
        out = backwards.as_bytes(fd.getvalue())
    else:
        fd = backwards.BytesIO()
        for ele in arr1:
            line = format_message(ele.tolist(), backwards.as_bytes(fmt_str))
            fd.write(line)
        # fmt = fmt_str.split(info['newline'])[0]
        # np.savetxt(fd, arr1,
        #            fmt=fmt, delimiter=info['delimiter'],
        #            newline=info['newline'], header='')
        out = fd.getvalue()
    fd.close()
    return out


def table_to_array(msg, fmt_str=None, use_astropy=False, names=None,
                   delimiter=None, comment=None, encoding='utf-8'):
    r"""Extract information from an ASCII table as an array.

    Args:
        msg (bytes): ASCII table as bytes string.
        fmt_str (bytes): Format string that should be used to parse the table.
            If not provided, this will attempt to determine the types of columns
            based on their contents.
        use_astropy (bool, optional): If True, astropy will be used to parse
            the table if it is installed. Defaults to False.
        names (list, optional): Field names that should be used for the
            structured data type of the output array. If not provided, names
            are generated based on the order of the fields in the table.
        delimiter (str, optional): String used to separate columns. Defaults to
            None and is not used. This is only used if fmt_str is not provided.
        comment (str, optional): String used to denote comments. Defaults to
            None and is not used. This is only used if fmt_str is not provided.
        encoding (str, optional): Encoding that should be used in Python 3 or
            higher to extract information from the message. Defaults to 'utf-8'.

    Returns:
        np.ndarray: Table contents as an array.
    
    """
    if not _use_astropy:
        use_astropy = False
    if fmt_str is None:
        dtype = None
        info = dict(delimiter=delimiter, comment=comment)
    else:
        dtype = cformat2nptype(fmt_str, names=names)
        info = format2table(fmt_str)
        names = dtype.names
    fd = backwards.BytesIO(msg)
    if names is not None:
        names = [backwards.as_str(n) for n in names]
    np_kws = dict()
    if info.get('delimiter', None) is not None:
        np_kws['delimiter'] = info['delimiter']
    if info.get('comment', None) is not None:
        np_kws['comments'] = info['comment']
    for k, v in np_kws.items():
        np_kws[k] = backwards.as_str(v)
    if use_astropy:
        # fd = backwards.StringIO(backwards.as_str(msg))
        if 'comments' in np_kws:
            np_kws['comment'] = np_kws.pop('comments')
        tab = apy_ascii.read(fd, names=names, guess=True,
                             encoding=encoding,
                             format='no_header', **np_kws)
        arr = tab.as_array()
        typs = [arr.dtype[i].str for i in range(len(arr.dtype))]
        cols = [c for c in tab.columns]
        # Convert type bytes if python 3
        if not backwards.PY2:  # pragma: Python 3
            new_typs = copy.copy(typs)
            convert = []
            for i in range(len(arr.dtype)):
                if np.issubdtype(arr.dtype[i], np.dtype('U')):
                    new_typs[i] = 'S' + typs[i].split('U')[-1]
                    convert.append(i)
            if convert:
                old_arr = arr
                new_dtyp = np.dtype([(c, t) for c, t in zip(cols, new_typs)])
                new_arr = np.zeros(arr.shape, new_dtyp)
                for i in range(len(arr.dtype)):
                    if i in convert:
                        x = np.char.encode(old_arr[cols[i]], encoding='utf-8')
                        new_arr[cols[i]] = x
                    else:
                        new_arr[cols[i]] = old_arr[cols[i]]
                arr = new_arr
                typs = new_typs
        # Convert complex type
        for i in range(len(arr.dtype)):
            if np.issubdtype(arr.dtype[i], np.dtype('S')):
                new_typs = copy.copy(typs)
                new_typs[i] = 'complex'
                new_dtyp = np.dtype([(c, t) for c, t in zip(cols, new_typs)])
                try:
                    arr = arr.astype(new_dtyp)
                except ValueError:
                    pass
        if dtype is not None:
            arr = arr.astype(dtype)
    else:
        np_ver = tuple([float(x) for x in (np.__version__).split('.')])
        if (np_ver >= (1.0, 14.0, 0.0)):
            arr = np.genfromtxt(fd, encoding='bytes', autostrip=True, dtype=None,
                                names=names, **np_kws)
        else:
            arr = np.genfromtxt(fd, autostrip=True, dtype=None,
                                names=names, **np_kws)
        if dtype is not None:
            arr = arr.astype(dtype)
    fd.close()
    return arr


def array_to_bytes(arrs, dtype=None, order='C'):
    r"""Serialize an array to bytes.

    Args:
        arrs (np.ndarray, list, tuple): Structured array or list/tuple of
            arrays that should be serialized. If 'arrs' is not a numpy
            structured array, a structured array will first be constructed
            from the provided data.
        dtype (np.dtype, optional): Structured data type that array should
            be converted to prior to serialization. If not provided, it
            will be determined from the array data.
        order (str, optional): Order that array should be serialized in.
            'C' for row first, 'F' for column/field first. Defaults to 'C'.

    Returns:
        bytes, str: Serialized array.

    Raises:

    """
    arr1 = consolidate_array(arrs, dtype=dtype)
    if order == 'F':
        ntyp = len(arr1.dtype)
        if ntyp == 0:
            out = arr1.tobytes(order='F')
        else:
            out = b''
            for i in range(ntyp):
                n = arr1.dtype.names[i]
                if np.issubdtype(arr1.dtype[i], np.dtype('complex')):
                    out = out + arr1[n].real.tobytes(order='F')
                    out = out + arr1[n].imag.tobytes(order='F')
                else:
                    out = out + arr1[n].tobytes(order='F')
    else:
        out = arr1.tobytes(order='C')
    return out


def bytes_to_array(data, dtype, order='C', shape=None):
    r"""Deserialize an array from bytes data.

    Args:
        data (bytes): Serialized array data.
        dtype (np.dtype, optional): Structured data type of the serialized array.
        order (str, optional): Order that array was serialized in.
            'C' for row first, 'F' for column/field first. Defaults to 'C'.
        shape (tuple, optional): Shape that array should be reshaped to
            after being deserialized. Defaults to a flat array if not provided.

    Returns:
        np.ndarray: Deserialized array.

    Raises:

    """
    if (len(data) % dtype.itemsize) != 0:
        raise RuntimeError("Data length (%d) is not a multiple " % len(data)
                           + "of the itemsize (%d)." % dtype.itemsize)
    nele = len(data) // dtype.itemsize
    ntyp = len(dtype)
    if (order == 'F') and (ntyp > 0):
        arr = np.empty((nele,), dtype=dtype)
        prev = 0
        j = 0
        for i in range(ntyp):
            idata = data[prev:(prev + (nele * dtype[i].itemsize))]
            if np.issubdtype(dtype[i], np.dtype('complex')):
                idata_real = idata[:(nele * dtype[i].itemsize // 2)]
                idata_imag = idata[(nele * dtype[i].itemsize // 2):]
                arr_real = np.frombuffer(idata_real, dtype='float64')
                arr_imag = np.frombuffer(idata_imag, dtype='float64')
                # arr_real = np.fromstring(idata_real, dtype='float64')
                # arr_imag = np.fromstring(idata_imag, dtype='float64')
                arr[dtype.names[i]] = np.zeros(nele, dtype=dtype[i])
                arr[dtype.names[i]] += arr_real
                arr[dtype.names[i]] += arr_imag * 1j
            else:
                arr[dtype.names[i]] = np.frombuffer(idata, dtype=dtype[i])
                # arr[dtype.names[i]] = np.fromstring(idata, dtype=dtype[i])
            prev += len(idata)
            j += 1
    else:
        arr = np.frombuffer(data, dtype=dtype)
        # arr = np.fromstring(data, dtype=dtype)
    # Reshape array
    if shape is not None:
        arr = arr.reshape(shape, order=order)
    return arr


def format_header(format_str=None, dtype=None,
                  comment=None, delimiter=None, newline=None,
                  field_names=None, field_units=None):
    r"""Get header lines for a table based on table information.

    Args:
        format_str (bytes, str, optional): Format string describing how the
            table should be formatted. If not provided, information on the
            formats is extracted from dtype.
        dtype (np.dtype, optional): Structured data type specifying the types
            of fields in the table. If not provided and format_str not
            specified, the formats will not be part of the header.
        comment (bytes, optional): String that should be used to comment the
            header lines. If not provided and not in format_str, defaults to
            _default_comment.
        delimiter (bytes, optional): String that should be used to separate
            columns. If not provided and not in format_str, defaults to
            _default_delimiter.
        newline (bytes, optional): String that should be used to end lines in
            the table. If not provided and not in format_str, defaults to
            _default_newline.
        field_names (list, optional): List of field names that should be
            included in the header. If not provided and dtype is None, names
            will not be included in the header.
        field_units (list, optional): List of field units that should be
            included in the header. If not provided, units will not be
            included.

    Returns:
        bytes: Bytes lines comprising a table header.

    Raises:
        ValueError: If there are not any format, names or units specified.

    """
    # Set defaults
    fmts = None
    if format_str is not None:
        info = format2table(format_str)
        if len(info['fmts']) > 0:
            fmts = info['fmts']
            if delimiter is None:
                delimiter = info['delimiter']
            if newline is None:
                newline = info.get('newline', None)
            if comment is None:
                comment = info.get('comment', None)
    if dtype is not None:
        if fmts is None:
            fmts = nptype2cformat(dtype, asbytes=True)
        if field_names is None:
            field_names = [backwards.as_bytes(n) for n in dtype.names]
    if delimiter is None:
        delimiter = _default_delimiter
    if comment is None:
        comment = _default_comment
    if newline is None:
        newline = _default_newline
    # Get count of fields
    if fmts is not None:
        nfld = len(fmts)
    elif field_names is not None:
        nfld = len(field_names)
    elif field_units is not None:
        nfld = len(field_units)
    else:
        raise ValueError("No formats, names, or units provided.")
    # Create lines
    out = []
    for x in [field_names, field_units, fmts]:
        if (x is not None) and (len(max(x, key=len)) > 0):
            assert(len(x) == nfld)
            out.append(comment
                       + delimiter.join([backwards.as_bytes(ix) for ix in x]))
    out = newline.join(out) + newline
    return out


def discover_header(fd, serializer, newline=_default_newline,
                    comment=_default_comment, delimiter=None,
                    lineno_format=None, lineno_names=None, lineno_units=None,
                    use_astropy=False):
    r"""Discover ASCII table header info from a file.

    Args:
        fd (file): File object containing the table.
        serializer (DefaultSerialize): Serializer that should be updated with
            header information.
        newline (str, optional): Newline character that should be used to split
            header if it is not already a list. Defaults to _default_newline.
        comment (bytes, optional): String that should be used to mark the
            header lines. If not provided and not in format_str, defaults to
            _default_comment.
        delimiter (bytes, optional): String that should be used to separate
            columns. If not provided and not in format_str, defaults to
            _default_delimiter.
        lineno_format (int, optional): Line number where formats are located.
            If not provided, an attempt will be made to locate one.
        lineno_names (int, optional): Line number where field names are located.
            If not provided, an attempt will be made to locate one.
        lineno_units (int, optional): Line number where field units are located.
            If not provided, an attempt will be made to locate one.
        use_astropy (bool, optional): If True, astropy will be used to parse
            the table if it is installed. Defaults to False.

    """
    header_lines = []
    header_size = 0
    prev_pos = fd.tell()
    for line in fd:
        sline = backwards.as_bytes(line.replace(
            backwards.as_bytes(platform._newline), newline))
        if not sline.startswith(comment):
            break
        header_size += len(line)
        header_lines.append(sline)
    # Parse header & set serializer attributes
    header = parse_header(header_lines, newline=newline,
                          lineno_format=lineno_format,
                          lineno_names=lineno_names,
                          lineno_units=lineno_units)
    # Override header with information set explicitly in serializer
    for k in serializer._oldstyle_kws:
        v = getattr(serializer, k, None)
        if v is not None:
            header[k] = v
    header.setdefault('format_str', None)
    if (delimiter is None) or ('format_str' in header):
        delimiter = header['delimiter']
    # Try to determine format from array without header
    str_fmt = b'%s'
    if ((header['format_str'] is None) or (str_fmt in header['format_str'])):
        fd.seek(prev_pos + header_size)
        all_contents = fd.read()
        if len(all_contents) == 0:  # pragma: debug
            return  # In case the file has not been written
        arr = table_to_array(all_contents,
                             names=header.get('field_names', None),
                             comment=comment,
                             delimiter=delimiter,
                             use_astropy=use_astropy)
        header['field_names'] = arr.dtype.names
        # Get format from array
        if header['format_str'] is None:
            header['format_str'] = table2format(
                arr.dtype, delimiter=delimiter,
                comment=b'',
                newline=header['newline'])
        # Determine maximum size of string field
        while str_fmt in header['format_str']:
            field_formats = extract_formats(header['format_str'])
            ifld = backwards.as_str(
                header['field_names'][field_formats.index(str_fmt)])
            max_len = len(max(arr[ifld], key=len))
            new_str_fmt = backwards.as_bytes('%' + str(max_len) + 's')
            header['format_str'] = header['format_str'].replace(
                str_fmt, new_str_fmt, 1)
    # Update serializer
    serializer.initialize_serializer(header)
    # Seek to just after the header
    fd.seek(prev_pos + header_size)


def parse_header(header, newline=_default_newline, lineno_format=None,
                 lineno_names=None, lineno_units=None):
    r"""Parse an ASCII table header to get information about the table.

    Args:
        header (list, str): Header lines that should be parsed.
        newline (str, optional): Newline character that should be used to split
            header if it is not already a list. Defaults to _default_newline.
        lineno_format (int, optional): Line number where formats are located.
            If not provided, an attempt will be made to locate one.
        lineno_names (int, optional): Line number where field names are located.
            If not provided, an attempt will be made to locate one.
        lineno_units (int, optional): Line number where field units are located.
            If not provided, an attempt will be made to locate one.

    Returns:
        dict: Parameters describing the information determined from the header.

    """
    out = dict()
    excl_lines = []
    if isinstance(header, backwards.bytes_type):
        header = [h + newline for h in header.split(newline)]
    # Locate format line
    if lineno_format is None:
        for i in range(len(header)):
            if len(extract_formats(header[i])) > 0:
                lineno_format = i
                break
    # Get format information from format line
    if lineno_format is not None:
        excl_lines.append(lineno_format)
        info = format2table(header[lineno_format])
        ncol = len(info['fmts'])
        out.update(**info)
        out['format_str'] = header[lineno_format].split(info['comment'])[-1]
    else:
        ncol = 0
        out.update(delimiter=_default_delimiter, comment=_default_comment,
                   fmts=[])
    out.setdefault('newline', newline)
    # Use explicit lines for names & units
    if lineno_names is not None:
        excl_lines.append(lineno_names)
        out['field_names'] = header[lineno_names].split(out['comment'])[-1].split(
            out['newline'])[0].split(out['delimiter'])
    if lineno_units is not None:
        excl_lines.append(lineno_units)
        out['field_units'] = header[lineno_units].split(out['comment'])[-1].split(
            out['newline'])[0].split(out['delimiter'])
    # Locate units & names
    for i in range(len(header)):
        if ('field_units' in out) and ('field_names' in out):
            break
        if i in excl_lines:
            continue
        cols = header[i].split(out['comment'])[-1].split(
            out['newline'])[0].split(out['delimiter'])
        if (len(cols) == ncol) or (ncol == 0):
            ncol = len(cols)
            is_unit = True
            for u in cols:
                if not units.is_unit(u):
                    is_unit = False
                    break
            if is_unit:
                if 'field_units' in out:
                    raise RuntimeError("Two lines could contain the field units.")
                out['field_units'] = cols
            else:
                if 'field_names' in out:
                    raise RuntimeError("Two lines could contain the field names.")
                out['field_names'] = cols
    return out


def dict2list(d, order=None):
    r"""Covert a dictionary of arrays to a list of arrays.

    Args:
        d (dict): Dictionary of arrays.
        order (list, optional): Key order in which values from the dictionary
            should be add to the list.

    Returns:
        list: List with contents from the input array.

    """
    if not isinstance(d, dict):
        raise TypeError("d must be a dictionary, not %s." % type(d))
    if order is None:
        order = sorted(list(d.keys()))
    out = [d[k] for k in order]
    return out


def list2dict(l, names=None):
    r"""Convert a list of arrays to a dictionary of arrays.

    Args:
        l (list): List of arrays.
        names (list, optional): Names to give to the new fields. Defaults to
            names based on order (e.g. 'f0', 'f1').

    Returns:
        dict: Dictionary of arrays.

    """
    if not isinstance(l, (list, tuple)):
        raise TypeError("l must be a list or tuple, not %s." % type(l))
    if names is None:
        names = ['f%d' % i for i in range(len(l))]
    out = {k: x for k, x in zip(names, l)}
    return out


def numpy2list(arr):
    r"""Covert a numpy structured array to a list of arrays.

    Args:
        arr (np.ndarray): Array to convert.

    Returns:
        list: List with contents from the input array.

    """
    if not isinstance(arr, np.ndarray):
        raise TypeError("arr must be a numpy array, not %s." % type(arr))
    return dict2list(numpy2dict(arr), order=arr.dtype.names)


def list2numpy(l, names=None):
    r"""Convert a list of arrays to a numpy structured array.

    Args:
        l (list): List of arrays.
        names (list, optional): Names to give to the new fields. Defaults to
            names based on order.

    Returns:
        np.ndarray: Structured numpy array.

    """
    return dict2numpy(list2dict(l, names=names), order=names)


def numpy2dict(arr):
    r"""Covert a numpy structured array to a dictionary of arrays.

    Args:
        arr (np.ndarray): Array to convert.

    Returns:
        dict: Dictionary with contents from the input array.

    """
    if not isinstance(arr, np.ndarray):
        raise TypeError("arr must be a numpy array, not %s." % type(arr))
    out = dict()
    if arr.dtype.names is None:
        out['0'] = arr
    else:
        for n in arr.dtype.names:
            out[n] = arr[n]
    return out


def dict2numpy(d, order=None):
    r"""Convert a dictionary of arrays to a numpy structured array.

    Args:
        d (dict): Dictionary of arrays.
        order (list, optional): Order that keys should be in array.
            Defaults to None and will be sorted keys.

    Returns:
        np.ndarray: Structured numpy array.

    """
    if not isinstance(d, dict):
        raise TypeError("d must be a dictionary, not %s." % type(d))
    if len(d) == 0:
        return np.array([])
    if order is None:
        order = sorted([k for k in d.keys()])
    dtypes = [d[k].dtype for k in order]
    dtype = np.dtype(dict(names=order, formats=dtypes))
    shape = d[order[0]].shape
    out = np.empty(shape, dtype)
    for n in order:
        out[n] = d[n]
    return out


def numpy2pandas(arr):
    r"""Covert a numpy structured array to a Pandas DataFrame.

    Args:
        arr (np.ndarray): Array to convert.

    Returns:
        pandas.DataFrame: Pandas data frame with contents from the input array.

    """
    if not isinstance(arr, np.ndarray):
        raise TypeError("arr must be a numpy array, not %s." % type(arr))
    out = pandas.DataFrame(arr)
    return out


def pandas2numpy(frame, index=False):
    r"""Convert a Pandas DataFrame to a numpy structured array.

    Args:
        frame (pandas.DataFrame): Frame to convert.
        index (bool, optional): If True, the index will be included as a field
            in the array. Defaults to False.

    Returns:
        np.ndarray: Structured numpy array.

    """
    if not isinstance(frame, pandas.DataFrame):
        raise TypeError("frame must be a pandas data frame, not %s." % type(frame))
    if frame.empty:
        return np.array([])
    arr = frame.to_records(index=index)
    # Covert object type to string
    old_dtype = arr.dtype
    new_dtype = dict(names=old_dtype.names, formats=[])
    for i in range(len(old_dtype)):
        if old_dtype[i] == object:
            try:
                max_len = len(max(arr[old_dtype.names[i]], key=len))
                new_dtype['formats'].append(np.dtype("S%s" % max_len))
            except TypeError:
                new_dtype['formats'].append(old_dtype[i])
        else:
            new_dtype['formats'].append(old_dtype[i])
    # Convert to unstructured array if name is default and only one column
    if (len(new_dtype['names']) == 1) and (new_dtype['names'][0] == '0'):
        out = np.zeros(arr.shape, dtype=new_dtype['formats'][0])
        out[:] = arr[new_dtype['names'][0]]
    else:
        new_dtype = np.dtype(new_dtype)
        out = np.zeros(arr.shape, dtype=new_dtype)
        out[:] = arr[:]
    return out


def dict2pandas(d, order=None):
    r"""Convert a dictionary of arrays to a Pandas DataFrame.

    Args:
        d (dict): Dictionary of arrays.
        order (list, optional): Order that keys should be in array.
            Defaults to None and will be sorted keys.

    Returns:
        pandas.DataFrame: Pandas data frame with contents from the input dict.

    """
    return numpy2pandas(dict2numpy(d, order=order))


def pandas2dict(frame):
    r"""Convert a Pandas DataFrame to a numpy structured array.

    Args:
        frame (pandas.DataFrame): Frame to convert.

    Returns:
        dict: Dictionary with contents from the input frame.

    """
    if not isinstance(frame, pandas.DataFrame):
        raise TypeError("frame must be a pandas data frame, not %s." % type(frame))
    if frame.empty:
        return {}
    return numpy2dict(pandas2numpy(frame))


def list2pandas(l, names=None):
    r"""Convert a list of arrays to a Pandas DataFrame.

    Args:
        l (list): List of arrays.
        names (list, optional): Names to give to the new fields. Defaults to
            names based on order (e.g. 'f0', 'f1').

    Returns:
        pandas.DataFrame: Pandas data frame with contents from the input list.

    """
    return numpy2pandas(list2numpy(l, names=names))


def pandas2list(frame):
    r"""Convert a Pandas DataFrame to a list of arrays.

    Args:
        frame (pandas.DataFrame): Frame to convert.

    Returns:
        list: List with contents from the input frame.

    """
    if frame.empty:
        return []
    return numpy2list(pandas2numpy(frame))


__all__ = []



from yggdrasil import backwards
from yggdrasil.serialize import _default_newline
from yggdrasil.serialize.SerializeBase import SerializeBase
from yggdrasil.metaschema.datatypes.PlyMetaschemaType import PlyDict


class PlySerialize(SerializeBase):
    r"""Class for serializing/deserializing .ply file formats.

    Args:
        write_header (bool, optional): If True, headers will be added to
            serialized output. Defaults to True.
        newline (str, optional): String that should be used for new lines.
            Defaults to '\n'.

    Attributes:
        write_header (bool): If True, headers will be added to serialized
            output.
        newline (str): String that should be used for new lines.
        default_rgb (list): Default color in RGB that should be used for
            missing colors.

    """
    
    _seritype = 'ply'
    _schema_subtype_description = ('Serialize 3D structures using Ply format.')
    _schema_properties = {
        'newline': {'type': 'string',
                    'default': backwards.as_str(_default_newline)}}
    default_datatype = {'type': 'ply'}
    concats_as_str = False

    def __init__(self, *args, **kwargs):
        r"""Initialize immediately as default is only type."""
        super(PlySerialize, self).__init__(*args, **kwargs)
        self.initialized = True

    def func_serialize(self, args):
        r"""Serialize a message.

        Args:
            args: List of arguments to be formatted or numpy array to be
                serialized.

        Returns:
            bytes, str: Serialized message.

        """
        return backwards.as_bytes(self.datatype.encode_data(args, self.typedef))

    def func_deserialize(self, msg):
        r"""Deserialize a message.

        Args:
            msg: Message to be deserialized.

        Returns:
            obj: Deserialized message.

        """
        return PlyDict(self.datatype.decode_data(backwards.as_str(msg),
                                                 self.typedef))

    @classmethod
    def concatenate(cls, objects, **kwargs):
        r"""Concatenate objects to get object that would be recieved if
        the concatenated serialization were deserialized.

        Args:
            objects (list): Objects to be concatenated.
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            list: Set of objects that results from concatenating those provided.

        """
        if len(objects) == 0:
            return []
        total = objects[0]
        for x in objects[1:]:
            total = total.merge(x)
        return [total]
        
    @classmethod
    def get_testing_options(cls):
        r"""Method to return a dictionary of testing options for this class.

        Returns:
            dict: Dictionary of variables to use for testing.

        """
        out = super(PlySerialize, cls).get_testing_options()
        obj = PlyDict({'vertices': [{'x': float(0), 'y': float(0), 'z': float(0)},
                                    {'x': float(0), 'y': float(0), 'z': float(1)},
                                    {'x': float(0), 'y': float(1), 'z': float(1)}],
                       'faces': [{'vertex_index': [int(0), int(1), int(2)]}]})
        out.update(objects=[obj, obj],
                   empty=dict(vertices=[], faces=[]),
                   contents=(b'ply\n'
                             + b'format ascii 1.0\n'
                             + b'comment author ygg_auto\n'
                             + b'comment File generated by yggdrasil\n'
                             + b'element vertex 6\n'
                             + b'property double x\n'
                             + b'property double y\n'
                             + b'property double z\n'
                             + b'element face 2\nproperty list uchar int vertex_index\n'
                             + b'end_header\n'
                             + b'0.0000 0.0000 0.0000\n'
                             + b'0.0000 0.0000 1.0000\n'
                             + b'0.0000 1.0000 1.0000\n'
                             + b'0.0000 0.0000 0.0000\n'
                             + b'0.0000 0.0000 1.0000\n'
                             + b'0.0000 1.0000 1.0000\n'
                             + b'3 0 1 2\n'
                             + b'3 3 4 5\n'))
        out['concatenate'] = [([], [])]
        return out



from yggdrasil.metaschema import normalizer as normalizer_mod
from yggdrasil.metaschema.properties.MetaschemaProperty import MetaschemaProperty


class DefaultMetaschemaProperty(MetaschemaProperty):
    r"""Property class for 'default' property."""

    name = 'default'
    _replaces_existing = True
    _validate = False

    # @classmethod
    # def encode(cls, instance):
    #     r"""Encoder for the 'default' container property."""
    #     return instance

    # @classmethod
    # def validate(cls, validator, value, instance, schema):
    #     r"""Validation method for 'default' property."""
    #     return

    @classmethod
    def normalize(cls, normalizer, value, instance, schema):
        r"""Normalization method for 'default' property."""
        if (((not normalizer.NO_DEFAULTS)
             and isinstance(instance, normalizer_mod.UndefinedProperty))):
            return value
        return instance



import os
import pprint
import numpy as np
import shutil
import tempfile
import warnings
from yggdrasil import metaschema
from yggdrasil.tests import assert_raises, assert_equal


def test_func():  # pragma: debug
    pass


_valid_objects = {'unicode': u'hello',
                  'bytes': b'hello',
                  'float': float(1), 'int': int(1),
                  'uint': np.uint(1), 'complex': complex(1, 1),
                  '1darray': np.zeros(5), 'ndarray': np.zeros((5, 5)),
                  'object': {'a': 'hello'}, 'array': ['hello', 1],
                  'ply': {'vertices': [{k: 0.0 for k in 'xyz'},
                                       {k: 0.0 for k in 'xyz'},
                                       {k: 0.0 for k in 'xyz'}],
                          'faces': [{'vertex_index': [0, 1, 2]}]},
                  'obj': {'vertices': [{k: 0.0 for k in 'xyz'},
                                       {k: 0.0 for k in 'xyz'},
                                       {k: 0.0 for k in 'xyz'}],
                          'faces': [[{'vertex_index': 0},
                                     {'vertex_index': 1},
                                     {'vertex_index': 2}]]},
                  'schema': {'type': 'string'},
                  'function': test_func}


_normalize_objects = [
    ({'type': 'integer'}, '1', 1),
    ({'type': 'integer'}, 'hello', 'hello'),
    ({'type': 'number'}, '1.0', 1.0),
    ({'type': 'number'}, 'hello', 'hello'),
    ({'type': 'string'}, 1, '1'),
    ({'type': 'unicode'}, 1, u'1'),
    ({'type': 'bytes'}, 1, b'1'),
    ({'type': 'float'}, '1', 1.0),
    ({'type': 'float'}, 'hello', 'hello'),
    ({'type': 'int'}, '1', 1),
    ({'type': 'int'}, 'hello', 'hello'),
    ({'type': 'uint'}, '1', np.uint(1)),
    ({'type': 'uint'}, 'hello', 'hello'),
    ({'type': 'complex'}, '(1+0j)', (1 + 0j)),
    ({'type': 'complex'}, 'hello', 'hello'),
    ({'type': 'object',
      'properties': {'a': {'type': 'int', 'default': '2'},
                     'b': {'type': 'int', 'default': '2'}}},
     {'b': '2'}, {'a': 2, 'b': 2}),
    ({'type': 'object',
      'definitions': {'ab': {'type': 'int', 'default': '1'}},
      'properties': {'a': {'$ref': '#/definitions/ab'},
                     'b': {'$ref': '#/definitions/ab'}}},
     {'b': '1'}, {'a': 1, 'b': 1}),
    ({'type': 'object',
      'definitions': {'ab': {'type': 'int'}},
      'required': ['a', 'b'],
      'properties': {'a': {'$ref': '#/definitions/ab'},
                     'b': {'$ref': '#/definitions/ab'}}},
     {'b': '1'}, {'b': '1'}),
    ({'type': 'object', 'properties': {'a': {'type': 'int'}}}, None, None),
    ({'type': 'object', 'properties': {'a': {'type': 'int'}}}, {}, {}),
    ({'type': 'array', 'items': {'type': 'int'}},
     '1, 2', [1, 2]),
    ({'type': 'array', 'items': [{'type': 'int'}, {'type': 'int'}]},
     '1, 2', [1, 2]),
    ({'type': 'array', 'items': {'type': 'int'}}, None, None),
    ({'type': 'function'}, '%s:test_func' % __name__, test_func),
    ({'type': 'function'}, '%s:invalid_func' % __name__,
     '%s:invalid_func' % __name__),
    ({'type': 'schema'}, {'units': 'g'},
     {'type': 'scalar', 'units': 'g', 'subtype': 'float', 'precision': int(64)}),
    ({'type': 'any', 'temptype': {'type': 'float'}}, '1', float(1))]


def test_create_metaschema():
    r"""Test errors in create_metaschema."""
    assert(metaschema.get_metaschema())
    assert_raises(RuntimeError, metaschema.create_metaschema, overwrite=False)


def test_get_metaschema():
    r"""Test get_metaschema and ensure the metaschema is current."""
    temp = os.path.join(tempfile.gettempdir(), metaschema._metaschema_fbase)
    old_metaschema = metaschema.get_metaschema()
    try:
        shutil.move(metaschema._metaschema_fname, temp)
        metaschema._metaschema = None
        new_metaschema = metaschema.get_metaschema()
        new_id = new_metaschema.get('$id', new_metaschema.get('id', None))
        old_id = old_metaschema.get('$id', old_metaschema.get('id', None))
        assert(new_id is not None)
        assert(old_id is not None)
        if new_id != old_id:  # pragma: debug
            warnings.warn(("The locally generated metaschema would have a different "
                           "id than the default (%s vs. %s). Check that your "
                           "installation of jsonschema is up to date.") % (
                               new_id, old_id))
        else:
            try:
                assert_equal(new_metaschema, old_metaschema)
            except AssertionError:  # pragma: debug
                print("Old:\n%s" % pprint.pformat(old_metaschema))
                print("New:\n%s" % pprint.pformat(new_metaschema))
                raise
    except BaseException:  # pragma: debug
        shutil.move(temp, metaschema._metaschema_fname)
        raise
    shutil.move(temp, metaschema._metaschema_fname)


def test_get_validator():
    r"""Test get_validator."""
    metaschema.get_validator()


def test_validate_instance():
    r"""Test validate_instance."""
    for k, v in _valid_objects.items():
        metaschema.validate_instance(v, {'type': k})


def test_normalize_instance():
    r"""Test normalize_instance."""
    for schema, x, y in _normalize_objects:
        z = metaschema.normalize_instance(x, schema, test_attr=1,
                                          show_errors=(x != y))
        try:
            assert_equal(z, y)
        except BaseException:  # pragma: debug
            print(schema, x, y, z)
            raise
